

import re
# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w_]')


from skipole import skilift
from .. import utils

from skipole import FailPage, ValidateError, ServerError


def retrieve_managepage(skicall):
    "this call is for the manage special pages page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # clears any session data
    utils.clear_call_data(call_data)

    project = call_data['editedprojname']
    labeldict = skilift.labels(project)

    page_data[("adminhead","page_head","large_text")] = "Manage page labels"

    system_list = skilift.sys_list()
    page_data['system:col_label'] = system_list
    page_data['system:col_input'] = _make_list(project, system_list, labeldict)
    page_data['system:hidden_field1'] = system_list

    lib_list = skilift.lib_list()
    page_data['jq:col_label'] = lib_list
    page_data['jq:col_input'] = _make_list(project, lib_list, labeldict)
    page_data['jq:hidden_field1'] = lib_list

    user_label_list = [item for item in labeldict if ( (item not in system_list) and (item not in lib_list) )]
    if user_label_list:
        user_label_list.sort()
        page_data['user:col_label'] = user_label_list
        page_data['user:col_input'] = _make_list(project, user_label_list, labeldict)
        page_data['user:hidden_field1'] = user_label_list
    else:
        page_data['user:show'] = False



def _make_list(project, reflist, labeldict):
    "Creates a list of url's or string ident numbers of items in reflist"
    result = []
    for item in reflist:
        if item in labeldict:
            target = labeldict[item]
            if isinstance(target, str):
                # its a url
                result.append(target)
            else:
                # its a tuple
                if target[0] == project:
                    result.append(str(target[1]))
                else:
                    result.append(",".join(target))
        else:
            result.append('')
    return result


def submit_special_page(skicall):
    "Sets special page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    label = call_data["label"]
    ident_or_url = call_data["ident_or_url"]
    if (not label) or (not ident_or_url):
        raise FailPage(message = "label or ident missing")

    # check label is valid
    if _AN.search(label):
        raise FailPage(message = "The label can only contain A-Z, a-z, 0-9 and the underscore character.")

    if '_' in label:
        labelparts = label.split('_')
        for lpart in labelparts:
            if lpart.isdigit():
                raise FailPage(message = "Invalid label (Danger of confusion with a page ident, please avoid using digits without letters).")

    if label.isdigit():
        raise FailPage(message = "Invalid label (Danger of confusion with a page ident).")

    try:
        skilift.set_label(project, label, ident_or_url)
    except ServerError as e:
        raise FailPage(message = e.message)

    if label in skilift.sys_list():
        call_data['status'] = 'System special page %s set' % (label,)
        return
    if label in skilift.lib_list():
        call_data['status'] = 'Library special page %s set' % (label,)
        return
    call_data['status'] = 'Label %s set' % (label,)


def add_user_page(skicall):
    "Sets user defined label and target"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    label = call_data["label"]
    ident_or_url = call_data["ident_or_url"]
    if (not label) or (not ident_or_url):
        raise FailPage(message = "label or ident missing", widget='addlabel')

    # check label is valid
    if _AN.search(label):
        raise FailPage(message = "The label can only contain A-Z, a-z, 0-9 and the underscore character.")

    if '_' in label:
        labelparts = label.split('_')
        for lpart in labelparts:
            if lpart.isdigit():
                raise FailPage(message = "Invalid label (Danger of confusion with a page ident, please avoid using digits without letters).")

    if label.isdigit():
        raise FailPage(message = "Invalid label (Danger of confusion with a page ident).")

    if label in skilift.lib_list():
        raise FailPage(message = "Invalid label, reserved for system files.")
    if label in skilift.sys_list():
        raise FailPage(message = "Invalid label, reserved for system pages.")

    try:
        skilift.set_label(project, label, ident_or_url)
    except ServerError as e:
        raise FailPage(message = e.message)
    call_data['status'] = 'Label %s set' % (label,)


def submit_user_page(skicall):
    "Edits or deletes a special user page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'label' not in call_data:
        raise FailPage(message = "Invalid label.")
    label = call_data["label"]
    if label in skilift.lib_list():
        raise FailPage(message = "Invalid label.")
    if label in skilift.sys_list():
        raise FailPage(message = "Invalid label.")
    if ("edit" in call_data) and (call_data["edit"]):
        submit_special_page(skicall)
        return
    if ("delete" not in call_data) or (not call_data["delete"]):
        raise FailPage(message = "Invalid submission.")
    try:
        skilift.delete_label(project, label)
    except ServerError as e:
        raise FailPage(message = e.message)
    call_data['status'] = 'User label deleted'





#
# This script is meant to be run from the command line using
#
# python3 -m skipole mynewproj /path/to/projectfiles
#
#

import sys, os, re, shutil

from . import version

DESCRIPTION = """Usage is

python3 -m skipole mynewproj /path/to/projectfiles

Which creates a directory /path/to/projectfiles
containing three sub directories:

mynewproj - a subdirectory containing a new project
skis - a project serving needed javascript files, necessary for all projects
skiadmin - a project used to help develop 'mynewproj'

You should replace 'mynewproj' with your preferred name for a new project.
This is an optional argument. If not given then only the skis and skiadmin
projects will be created.

The path "/path/to/projectfiles" must be given, and is the path to a directory
where you will develop your project. Multiple projects can be created in one
'projectfiles' directory, or you could have multiple such directories holding
different projects.

If the directory already exists, any existing skiadmin and skis projects will
be overwritten, this facility is used to upgrade skis and skiadmin if a new
version of skipole is installed.

If mynewproj already exists in the directory, it will not be changed.
"""


# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')


args = sys.argv

if len(args) == 3:
    project_name = args[1]
    projectfiles = os.path.abspath(os.path.expanduser(args[2]))
    if _AN.search(project_name):
        print( "Error: Invalid project name, alphanumeric only")
        sys.exit(1)
    if '_' in project_name:
        print( "Error: Invalid project name, alphanumeric only (no underscore).")
        sys.exit(1)
    if (project_name == 'skis') or (project_name=='skiadmin'):
        print("Error: This project name is reserved")
        sys.exit(2)
elif len(args) == 2:
    if args[1] == "--version":
        print(version)
        sys.exit(0)
    project_name = None
    projectfiles = os.path.abspath(os.path.expanduser(args[1]))
else:
    print( "Invalid input. " + DESCRIPTION)
    sys.exit(3)


# get the location of the directories to be copied
template_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'template_projects')

if not os.path.isdir(template_directory):
    print("Error: Cannot find the template data for the skis and skiadmin projects")
    sys.exit(4)
template_skis_directory = os.path.join(template_directory, 'skis')
if not os.path.isdir(template_skis_directory):
    print("Error: Cannot find the template data for the skis project")
    sys.exit(5)
template_skiadmin_directory = os.path.join(template_directory, 'skiadmin')
if not os.path.isdir(template_skiadmin_directory):
    print("Error: Cannot find the template data for the skiadmin project")
    sys.exit(6)
template_newproj_directory = os.path.join(template_directory, 'newproj')
if not os.path.isdir(template_newproj_directory):
    print("Error: Cannot find the template data for the newproj project")
    sys.exit(7)



# If the given projectfiles directory does not exist
if not os.path.isdir(projectfiles):

    # create the directory by copying the template_directory
    shutil.copytree(template_directory, projectfiles)

    # new directory which has been created
    newproj_directory = os.path.join(projectfiles, 'newproj')

    if project_name is None:
        # remove the new project
        shutil.rmtree(newproj_directory)
        print("Project directory %s created" % (projectfiles,))
        sys.exit(0)

    # A project name is given, change the name of the newproj_directory to project_name
    project_directory = os.path.join(projectfiles, project_name)
    os.rename(newproj_directory, project_directory)


    new_code_path = os.path.join(project_directory, "code", project_name + ".py")
    print("Creating %s" % (new_code_path,))

    # change newproj.py to project_name.py, including the contents
    ## READ newproj.py
    newproj_file = os.path.join(project_directory, "code", "newproj.py")
    with open(newproj_file, "r") as tnf:
        templatecontents = tnf.read()
    os.remove(newproj_file)
    ## REPLACE newproj with the new project name
    newcontents = templatecontents.replace("newproj", project_name)
    ## WRITE the new file
    with open(new_code_path, "w") as pf:
        pf.write(newcontents)

    # change data/project.json contents
    ## READ project.json
    data_file = os.path.join(project_directory, "data", "project.json")
    with open(data_file, "r") as tdf:
        templatedatacontents = tdf.read()
    os.remove(data_file)
    ## REPLACE newproj with the new project name
    newdatacontents = templatedatacontents.replace("newproj", project_name)
    ## WRITE the new file
    with open(data_file, "w") as tdf2:
        tdf2.write(newdatacontents)


    print("Project directory %s created, with new project %s" % (projectfiles,project_name))
    print("""Use the command:
python3 %s
To serve the project at url localhost:8000
and localhost:8000/skiadmin for the admin pages""" % (new_code_path,))
    sys.exit(0)

# The given projectfiles does exist, so use it for the new project

print("Replacing project skis")
# if skis exists, delete it, then copy the template
skis_directory = os.path.join(projectfiles, 'skis')
if os.path.isdir(skis_directory):
    shutil.rmtree(skis_directory)
shutil.copytree(template_skis_directory, skis_directory)

print("Replacing project skiadmin")
# if skiadmin exists, delete it, then copy the template
skiadmin_directory = os.path.join(projectfiles, 'skiadmin')
if os.path.isdir(skiadmin_directory):
    shutil.rmtree(skiadmin_directory)
shutil.copytree(template_skiadmin_directory, skiadmin_directory)

if project_name is None:
    print("skis and skiadmin have been replaced.")
    sys.exit(0)

project_directory = os.path.join(projectfiles, project_name)
if os.path.isdir(project_directory):
    print("skis and skiadmin have been replaced, however project % already exists, and has not been altered." % (project_name,))
    sys.exit(0)

# create a new project
shutil.copytree(template_newproj_directory, project_directory)

new_code_path = os.path.join(project_directory, "code", project_name + ".py")
print("Creating %s" % (new_code_path,))

# change newproj.py to project_name.py, including the contents
## READ newproj.py
newproj_file = os.path.join(project_directory, "code", "newproj.py")
with open(newproj_file, "r") as tnf:
    templatecontents = tnf.read()
os.remove(newproj_file)
## REPLACE newproj with the new project name
newcontents = templatecontents.replace("newproj", project_name)
## WRITE the new file
with open(new_code_path, "w") as pf:
    pf.write(newcontents)

# change data/project.json contents
## READ project.json
data_file = os.path.join(project_directory, "data", "project.json")
with open(data_file, "r") as tdf:
    templatedatacontents = tdf.read()
os.remove(data_file)
## REPLACE newproj with the new project name
newdatacontents = templatedatacontents.replace("newproj", project_name)
## WRITE the new file
with open(data_file, "w") as tdf2:
    tdf2.write(newdatacontents)

print("skis and skiadmin have been replaced, and new project %s created." % (project_name,))
print("""Use the command:
python3 %s
To serve the project at url localhost:8000
and localhost:8000/skiadmin for the admin pages""" % (new_code_path,))
sys.exit(0)






#######################################################################################
#
# TextBlocks are a mapping of the tuple (textref, language) to a block of text
#
# Various widgets that display TextBlocks are given a textref reference string, and
# the language is generally derived from a browser setting. The widget then obtains
# the block of text and displays it.
# 
# To obtain the text, the framework imports this module and creates an instance
# of the AccessTextBlocks object, and uses its methods to find the text.
#
# This class reads TextBlocks from JSON files, stores them into memory and
# serves them to the framework when requested.
#
# The class also has methods for writing and saving TextBlocks to the JSON files
# which are used by skiadmin to create TextBlocks.
#
########################################################################################


import os, json, shutil


class AccessTextBlocks(object):


    def __init__(self, project, projectfiles, default_language):
        """The project imports this module and creates an instance of this class, and uses it to read the
           text of TextBlocks.
           The text is stored in JSON files beneath the directory
           projectfiles/project/data/textblocks_json
        """

        # the self.default_language attribute is expected to exist
        self.default_language = default_language
        # as is a set of available languages
        self._languages = set()
 
        # this implementation holds text in memory in this dictionary with keys
        # of the tuple (reference,language) and values being the text
        self._textblocks = {}
        # dictionary of textrefs with keys of reference and values being a list of languages for that textref
        self._textrefs = {}
        # read the json files from this directory and populate the above two dictionaries and self._languages
        self._textblocks_json_directory = os.path.join(projectfiles, project, "data", "textblocks_json")
        # for each language file in the directory, add it to the dictionary
        dir_contents = os.listdir(self._textblocks_json_directory)
        for filename in dir_contents:
            if not filename.endswith('.json'):
                continue
            filepath = os.path.join(self._textblocks_json_directory,filename)
            if not os.path.isfile(filepath):
                continue
            # filenames refer to languages, such as en.json, de.json
            language = filename[:-5].lower()
            self._languages.update([language])
            with open(filepath, 'r') as fp:
                textblocks_dict = json.load(fp)
                for textref, text in textblocks_dict.items():
                   self._textblocks[(textref,language)] = text
                   if textref in self._textrefs:
                       if language not in self._textrefs[textref]:
                           self._textrefs[textref].append(language)
                   else:
                        self._textrefs[textref] = [language]
        # sort language lists in self._textrefs
        for langlist in self._textrefs.values():
            langlist.sort()


    @property
    def languages(self):
        "Property which is a set of available languages"
        return self._languages

    def get_textrefs(self):
        "Returns a list of the textrefs"
        return list(self._textrefs.keys())

    def get_textref_languages(self):
        "Return a dictionary {textref: [languages],...}"
        return self._textrefs.copy()

    def textref_exists(self, textref):
        "Return True if the textref exists"
        return textref in self._textrefs

    def get_exact_text(self, textref, language):
        "Get text with given textref and language, gets exact value, does not seek nearest, if not found return None"
        if (textref,language) in self._textblocks:
            return self._textblocks[(textref,language)]

    def get_text(self, textref, lang):
        """Gets the text from the textblock, trying nearest language, returns None if not found
           lang is a tuple of preferred language, default_language"""
        if textref not in self._textrefs:
            return
        language, default_language = lang
        language = language.lower()
        default_language = default_language.lower()
        # try preferred language
        if (textref,language) in self._textblocks:
            return self.get_exact_text(textref,language)
        # if not preferred for example en-gb, try en
        shortlang = language.split('-')
        if (textref,shortlang[0]) in self._textblocks:
            return self.get_exact_text(textref,shortlang[0])
        # try default language
        if (textref,default_language) in self._textblocks:
            return self.get_exact_text(textref,default_language)
        shortlang = default_language.split('-')
        if (textref,shortlang[0]) in self._textblocks:
            return self.get_exact_text(textref,shortlang[0])
        # try self.default_language
        if (textref,self.default_language) in self._textblocks:
            return self.get_exact_text(textref,self.default_language)
        shortlang = self.default_language.split('-')
        if (textref,shortlang[0]) in self._textblocks:
            return self.get_exact_text(textref,shortlang[0])
        # try whatever language remains
        rlang = self.textrefs[textref][0]
        return self.get_exact_text(textref,rlang)

    def set_text(self, text, textref, language):
        """Sets the text into the textblock, with the given textref and language
           Used by skiadmin, in this implementation it is put into memory but
           not saved to the json files until the save method is called when the project
           is committed"""
        language = language.lower()
        if not text:
            self.del_text(textref, language)
            return
        self._textblocks[textref, language] = text
        self._languages.update([language])
        if textref in self._textrefs:
            langlist = self._textrefs[textref]
            if language not in langlist:
                langlist.append(language)
                langlist.sort()
        else:
            self._textrefs[textref] = [language]
        # if the language is of the format 'en-gb' then check if en is present, if not, add it
        shortlang = language.split('-')
        if shortlang[0] == language:
            return
        if (textref, shortlang) in self._textblocks:
            # A textblock with the given shortlang does exist, but check to see if it is just an empty string
            shortlangtext = self._textblocks[(textref, shortlang)]
            if (not shortlangtext) or (shortlangtext == " "):
                # add text
                self.set_text(text, textref, shortlang)
        else:
            # shortlang does not exist, so add it
            self.set_text(text, textref, shortlang)

    def del_text(self, textref, language):
        "Deletes the text from the textblock with the given language"
        language = language.lower()
        if (textref,language) not in self._textblocks:
            return
        del self._textblocks[(textref,language)]
        langlist = self._textrefs[textref]
        try:
            pos = langlist.index(language)
            del langlist[pos]
            if not langlist:
                del self._textrefs[textref]
        except Exception:
            pass
        # test if language used anywhere
        for langlist in self._textrefs.values():
            if language in langlist:
                break
        else:
            self._languages.discard(language)

    def del_textblock(self, textref):
        "Deletes the textblock, all languages"
        langlist = self._textrefs[textref].copy()
        for language in langlist:
            self.del_text(textref, language)

    def copy(self, sourceref, destinationref):
        "Copies textblock"
        langlist = self._textrefs[sourceref]
        self._textrefs[destinationref] = []
        destlanglist = self._textrefs[destinationref]
        for language in langlist:
            self._textblocks[(destinationref, language)] = self._textblocks[(sourceref, language)]
            destlanglist.append(language)

    def save(self):
        "Creates json files from the textblocks"
        if os.path.isdir(self._textblocks_json_directory):
            # remove existing json files
            shutil.rmtree(self._textblocks_json_directory)
        os.mkdir(self._textblocks_json_directory)
        for language in self._languages:
            filepath = os.path.join(self._textblocks_json_directory, language + '.json')
            lang = (language, self.default_language)
            json_dict = {}
            for textref, langlist in self._textrefs.items():
                if language not in langlist:
                    continue
                json_dict[textref] = ''
                text = self.get_text(textref, lang)
                if text is not None:
                    json_dict[textref] = text
            # save the json dictionary to file
            with open(filepath, 'w') as fp:
                json.dump(json_dict, fp, indent=2, sort_keys=True)






import os, sys

# This project needs to import the skipole package, which should normally be immediately
# available if skipole has been installed on your system.

# from skipole import the WSGIApplication class which will be used to create a wsgi
# application, and exception classes which you will need in your functions

from skipole import WSGIApplication, FailPage, GoTo, ValidateError, ServerError, set_debug

# The set_debug function is used during development, it adds exception data to server error
# pages, and also stops javascript validation of input data on the client - this ensures that
# server validation can be tested

# You may also wish to import 'use_submit_list' whch can act as a decorator around your
# submit_data function. See the skipole documentation for details.

# the framework needs to know the location of the projectfiles directory holding this and
# other projects - specifically the skis and skiadmin projects
# The following line assumes, as default, that this script file is located beneath
# ...projectfiles/newproj/code/

PROJECTFILES = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
PROJECT = 'newproj'



##############################################################################
#
# Your code needs to provide your own version of the following three functions
#
# Minimal versions are provided below, you could either develop them here,
# or more usually, create further modules and packages and import them.
#
##############################################################################


def start_call(called_ident, skicall):
    "When a call is initially received this function is called."
    return called_ident


def submit_data(skicall):
    "This function is called when a Responder wishes to submit data for processing in some manner"
    return


def end_call(page_ident, page_type, skicall):
    """This function is called at the end of a call prior to filling the returned page with skicall.page_data,
       it can also return an optional session cookie string."""
    return


##############################################################################
#
# The above functions will be inserted into the skipole.WSGIApplication object
# and will be called as required
#
##############################################################################


# create the wsgi application
application = WSGIApplication(project=PROJECT,
                              projectfiles=PROJECTFILES,
                              proj_data={},
                              start_call=start_call,
                              submit_data=submit_data,
                              end_call=end_call,
                              url="/")

# This creates a WSGI application object. On being created the object uses the projectfiles location to find
# and load json files which define the project, and also uses the functions :
#     start_call, submit_data, end_call
# to populate returned pages.
# proj_data is an optional dictionary which you may use for your own purposes,
# it is included as the skicall.proj_data attribute


# The skis application must always be added, without skis you're going nowhere!
# The skis sub project serves javascript files required by the framework widgets.

skis_code = os.path.join(PROJECTFILES, 'skis', 'code')
if skis_code not in sys.path:
    sys.path.append(skis_code)
import skis
skis_application = skis.makeapp(PROJECTFILES)
application.add_project(skis_application, url='/lib')

# The add_project method of application, enables the added sub application
# to be served at a URL which should extend the URL of the main 'root' application.
# The above shows the main newproj application served at "/" and the skis library
# project served at "/lib"

# Note if you want to add further sub-projects you would typically:
#     Place the sub project code location on your sys.path
#     Import the sub project to obtain its wsgi application
#     Call application.add_project with the sub project application
#     and the url where it will be served.


# to deploy on a web server, you would typically install skipole on the web server,
# together with a 'projectfiles' directory containing the projects you want
# to serve (typically this project, and the skis project).
# you would then follow the web servers own documentation which should describe how
# to load a wsgi application.

# for example, using gunicorn3 by command line

# gunicorn3 -w 4 newproj:application

# Where gunicorn3 is the python3 version of gunicorn

#############################################################################
#
# You could remove everything below here when deploying and serving your
# finished application. The following lines are used to serve the project
# locally and add the skiadmin project for development.
#
#############################################################################

if __name__ == "__main__":


    ############################### THESE LINES ADD SKIADMIN FOR DEVELOPMENT ONLY #
    ###################### AND SHOULD BE REMOVED WHEN YOU DEPLOY YOUR APPLICATION #
                                                                                  #
    set_debug(True)                                                               #
    skiadmin_code = os.path.join(PROJECTFILES, 'skiadmin', 'code')                #
    if skiadmin_code not in sys.path:                                             #
        sys.path.append(skiadmin_code)                                            #
    import skiadmin                                                               #
    skiadmin_application = skiadmin.makeapp(PROJECTFILES, editedprojname=PROJECT) #
    application.add_project(skiadmin_application, url='/skiadmin')                #
                                                                                  #
    ###############################################################################

    # TYPICALLY WHEN YOUR FINISHED APPLICATION IS DEPLOYED, YOUR WEB SERVER WILL
    # IMPORT THIS MODULE AND THIS CODE WILL NEVER RUN AND COULD BE DELETED.
    # ALTERNATIVELY YOU MAY WISH TO ALTER THIS TO SERVE A DIFFERENT WEB SERVER
    # FROM THIS SCRIPT. FOR EXAMPLE, USING THE PYTHON 'WAITRESS' WEB SERVER:
    #
    # from waitress import serve
    # serve(application, host='0.0.0.0', port=8000)
    #

    from wsgiref.simple_server import make_server

    # serve the application
    host = "127.0.0.1"
    port = 8000

    httpd = make_server(host, port, application)
    print("Serving %s on port %s. Call http://localhost:%s/skiadmin to edit." % (PROJECT, port, port))
    httpd.serve_forever()






"""
This module defines the Folder and RootFolder classes.

A web site starts with one Folder instance as the root, and this in turn holds other Folder
instances (in a folders attribute dictionary), and Page instances (in a pages attribute dictionary).

New Folder instances which are sub folders of root should only be created
from a parent folder using the parents method *create_folder* or *create_admin_folder*

New Page instances should be similarly be created using method *create_page*
or *create_admin_page*
"""


import copy, os, cgi, collections, html, pprint, json, shutil, uuid

from http import cookies

from . import skiboot, responders
from .excepts import ServerError


class Folder(object):
    """Represents a folder as shown within a url

    A Folder instance contains items with keys corresponding to the folder and page names within the
    folder. so if fldr is a folder instance, and 'mypage' is a page name of a page under fldr:

    fldr['mypage'] is the corresponding Page instance. 

    *instance attributes:*

    *default_page_name*
      Normally "index", and is the page shown if the url only contains the folder name

    *folders*
      A dictionary of {folder_name: ident} for each folder in the folder

    *ident*
      A numeric number acting as a folder identifier, the root folder has number 1
      and further folders and pages are added with incrementing numbers.
      If this is an admin folder, the ident is a negative number.

    *name*
      The name of the folder, making up the url, and therefore should be lower case
      and url friendly

    *pages*
      A dictionary of {page_name: ident} for each page in the folder

    *parentfolder*
      This is the parent folder of this instance, the root folder has parentfolder None

    """

    # though this is not a page_type, it has a page_type class attribute so this
    # can be checked for any item
    page_type = "Folder"


    def __init__(self, name=None, brief = "New Folder", default_page_name = "", restricted=False):
        """Initiates a Folder instance

        name: None if root, any url friendly name for any other folder
        default_page_name: the page to be shown if the url gives this folder only
        """
        # self.pages is a dictionary of page.name:page.ident
        self.pages = {}
        # self.folders is a dictionary of folder.name:folder.ident
        self.folders = {}
        self.ident = None
        self._parentfolder_ident = None
        if name:
            self._name = name.lower()
        else:
            self._name = ""
        if default_page_name:
            self.default_page_name = default_page_name.lower()
        else:
            self.default_page_name = ""
        self._restricted = bool(restricted)
        self.brief = brief
        # the change is a uuid which alters whenever the page changes
        self.change = uuid.uuid4().hex

    @property
    def proj_ident(self):
        return self.ident.proj

    @property
    def project(self):
        if self.ident:
            return skiboot.getproject(proj_ident=self.ident.proj) 

    @property
    def restricted(self):
        return self._restricted

    def set_unrestricted(self):
        """Sets this folder as unrestricted, returns True if folder
           is unrestricted, or False (due to parent being restricted) if not"""
        if not self._restricted:
            # Folder is already unrestricted
            return True
        if self._parentfolder_ident is not None:
            parentfolder = skiboot.get_item(self._parentfolder_ident)
            if parentfolder.restricted:
                self._restricted = True
                # cannot set unrestricted as parent is restricted
                return False
        self._restricted = False
        return True


    def set_restricted(self):
        """Sets this folder as restricted, sets sub folders to also be restricted,
           returns a list of those folders (starting with self) which have been set to restricted,
           any that are already set, are not included in the list, so returns an empty list
           if this folder is already restricted """
        if self._restricted:
            # already restricted
            return []
        self._restricted = True
        self.default_page_name = ""
        restricted_list = [self]
        # set any subfolders to be restricted
        for folder_ident in self.folders.values():
            folder = skiboot.get_item(folder_ident)
            restricted_list.extend(folder.set_restricted())
        return restricted_list


    @property
    def default_page(self):
        "return a deep copy of the default page, or None if one not set"
        default_ident = self.default_page_ident()
        if default_ident:
            return skiboot.from_ident(default_ident)


    @property
    def default_page_ident(self):
        "return the default page ident, or None if one not set"
        if self._restricted:
            return
        if self.default_page_name and (self.default_page_name in self.pages):
            return self.pages[self.default_page_name]


    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = name.lower()

    name = property(get_name, set_name)

    def get_parentfolder(self):
        if self._parentfolder_ident is None:
            return None
        return skiboot.get_item(self._parentfolder_ident)

    def set_parentfolder(self, parentfolder):
        "Warning, this does not set this folder into the parents folders dictionary, it just sets this parentfolder attribute"
        if parentfolder is None:
            self._parentfolder_ident = None
        else:
            self._parentfolder_ident = parentfolder.ident

    parentfolder = property(get_parentfolder, set_parentfolder)

    @property
    def parentfolder_ident(self):
        "Returns the parentfolder ident"
        return self._parentfolder_ident


    def parent_list(self):
        "returns a list of (name, identnumber) starting at root"
        if self.ident.num == 0:
            return [('root', 0)]
        p_list = self.parentfolder.parent_list()
        p_list.append((self.name, self.ident.num))
        return p_list


    @property
    def url(self):
        return self._parentfolder_ident.url() + self.name + "/"


    def page_ident_from_path(self, identitems, pathlist):
        """"Return a page ident, where pathlist is a list of folder names and a final page name
        pathlist does not include the initial project url or this folders name
        identitems is the project {ident:folder or page} dictionary
        Return None if page not found, or a folder in the path is restricted"""
        if self._restricted:
            return
        # If pathlist is empty, return default page ident
        if not pathlist:
            return self.default_page_ident
        name = pathlist[0]
        # could be a page
        if name in self.pages:
            if len(pathlist) > 1:
                # page followed by further items is an error
                return
            # its a  page
            return self.pages[name]
        if name in self.folders:
            # requesting a folder
            folder_ident = self.folders[name]
            folder = identitems[folder_ident]
            return folder.page_ident_from_path(identitems, pathlist[1:])


    def ident_from_path(self, path):
        "Return the page or folder ident, or if not found, return None"
        myurl = self.url
        if (myurl == path) or (myurl == path+"/"):
            # request matches url path, therefore return self.ident
            return self.ident
        first_bit, name = path.rsplit("/", 1)
        path_string = first_bit + "/"
        if path_string == myurl:
            # this is a request for a page or a folder inside this folder
            if name in self.pages:
                # get the page
                return self.pages[name]
            elif name in self.folders:
                # requesting a folder
                return self.folders[name]
            else:
                # the page is not found
                return None
        elif path.find(myurl) == 0 :
            # The request may be for a sub folder of this page
            remaining_path = path[len(myurl):]
            bits = remaining_path.split("/", 1)
            name = bits[0]
            if name in self.folders:
                folder = skiboot.get_item(self.folders[name])
                return folder.ident_from_path(path)
            else:
                # the path is not found
                return None
        else:
            # the path is not found
            return None


    def add_page(self, page, ident=None):
        """Adds the page to this folder and project.
           The given ident can be None, Ident object, string or integer.
           Checks the given item is a page, respondpage etc .. (not a Folder) then calls project.add_item
           Returns the page ident."""
        if page.page_type == 'Folder':
            raise ServerError(message="Sorry, requested item is not a page")
        return self.project.add_item(self.ident, page, ident)


    def add_folder(self, folder, ident=None):
        """Adds the given folder to this folder and project.
           The given ident can be None, Ident object, string or integer.
           Checks given item is a Folder, then calls project.add_item
           Returns the folder ident"""
        if folder.page_type != 'Folder':
            raise ServerError(message="Sorry, requested item is not a folder")
        return self.project.add_item(self.ident, folder, ident)


    def folder_idents(self):
        "Returns a list of folder idents within this folder, sorted by name"
        name_list = [name for name in self.folders]
        name_list.sort()
        return [self.folders[name] for name in name_list]

    def page_idents(self):
        "Returns a list of page idents within this folder, sorted by name"
        name_list = [name for name in self.pages]
        name_list.sort()
        return [self.pages[name] for name in name_list]

    @property
    def ident_numbers(self):
        "return a list of ident numbers within this folder, not including folder number itself"
        num_list = []
        for ident in self.folders.values():
            num_list.append(ident.num)
        for ident in self.pages.values():
            num_list.append(ident.num)
        num_list.sort()
        return num_list

    def contains_ident(self, ident):
        "Returns True if the given ident is beneath this Folder"
        ident = skiboot.Ident.to_ident(ident, self.ident.proj)
        if ident in self.pages.values():
            return True
        if ident in self.folders.values():
            return True
        return False

    def __getitem__(self, name):
        "returns the page or folder in this folder - if it is not in the folder, returns None"
        if name in self.pages:
            page = skiboot.get_item(self.pages[name])
            if page is None:
                del self.pages[name]
            else:
                return page
        if name in self.folders:
            folder = skiboot.get_item(self.folders[name])
            if folder is None:
                del self.folders[name]
            else:
                return folder

    def __contains__(self, name):
        if name in self.pages:
            return True
        if name in self.folders:
            return True
        return False

    def __len__(self):
        return len(self.pages) + len(self.folders)

    def __bool__(self):
        return True

    def __eq__(self, other):
        if not isinstance(other, self.__class__): return False
        return self.ident == other.ident

    def __ne__(self, other):
        if not isinstance(other, self.__class__): return True
        return self.ident != other.ident

    def __str__(self):
        return self.url

    def __repr__(self):
        if self.ident:
            return "Folder ident %s" % (self.ident,)
        else:
            return "Folder name %s" % (self._name,)




class RootFolder(Folder):
    """Represents the site root - inherits from Folder with some methods
       overwritten"""

    def __init__(self, proj_ident, brief = "The site root folder", default_page_name = "index"):
        """Initiates a RootFolder instance"""
        Folder.__init__(self, name=None, brief=brief, default_page_name = default_page_name, restricted=False)
        self.ident = skiboot.Ident(proj_ident, 0)  # rootfolder always has Ident.num of zero
        self._restricted = False

    @property
    def restricted(self):
        # root folder cannot be a restricted folder
        return False

    def set_unrestricted(self):
        # root is always unrestricted
        return True

    def set_restricted(self):
        # restricted attribute cannot be set on the root
        return []


    def get_name(self):
        return None

    def set_name(self, name):
        raise ServerError(message="Sorry, Root Folder name cannot be set")

    name = property(get_name, set_name)

    def get_parentfolder(self):
        return None

    def set_parentfolder(self, parent):
        raise ServerError(message="Sorry, The Root Folder cannot have a parentfolder set")

    parentfolder = property(get_parentfolder, set_parentfolder)


    @property
    def url(self):
        "Returns the project url"
        folder_project = skiboot.getproject(self.ident.proj)
        return folder_project.url

    def __repr__(self):
        return "RootFolder ident %s" % (self.ident,)







"Functions to edit and insert a section placeholder"

import re

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w_]')

from skipole import FailPage, ValidateError, GoTo, ServerError
from skipole.skilift import part_info, editsection, item_info


def retrieve_editplaceholder(skicall):
    "Fills in the edit placeholder page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # a skilift.part_tuple is (project, pagenumber, page_part, section_name, name, location, part_type, brief)

    if 'part_tuple' in call_data:
        part_tuple = call_data['part_tuple']
        project = part_tuple.project
        pagenumber = part_tuple.pagenumber
        location = part_tuple.location
    else:
        project = call_data['editedprojname']
        pagenumber = call_data['page_number']
        location = call_data['location']
 
    # Fill in header
    page_data[("adminhead","page_head","large_text")] = "Edit Section place holder"

    # header done, now page contents

    # get placholderinfo
    placeholder = editsection.placeholder_info(project, pagenumber, location)
    if placeholder is None:
        raise FailPage("Section Place Holder not found")

    # get current sections
    section_list = editsection.list_section_names(project)

    if not section_list:
        page_data[('editsection','option_list')] = ['-None-']
        page_data[('editsection','selectvalue')] = '-None-'
        page_data[("nosection", "show")] = True
    elif placeholder.section_name in section_list:
        page_data[('editsection','option_list')] = section_list[:]
        page_data[('editsection','selectvalue')] = placeholder.section_name
    else:
        page_data[('editsection','option_list')] = ['-None-'] + section_list
        page_data[('editsection','selectvalue')] = '-None-'
        page_data[("nosection", "show")] = True

    page_data[("brief", "input_text")] = placeholder.brief
    page_data[("placename", "input_text")] = placeholder.alias
    page_data[("multiplier", "input_text")] = str(placeholder.multiplier)
    page_data[("mtag", "input_text")] = placeholder.mtag

    # set session data
    call_data['location'] = location


def set_placeholder(skicall):
    "Sets the placeholder section name, brief, or placename"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    location = call_data['location']
    pchange = call_data['pchange']

    call_data['part_tuple'] = part_info(project, pagenumber, None, location)

    # get current sections
    section_list = editsection.list_section_names(project)

    # get placholderinfo
    placeholder = editsection.placeholder_info(project, pagenumber, location)

    section_name = placeholder.section_name
    alias = placeholder.alias
    brief = placeholder.brief
    multiplier = placeholder.multiplier
    mtag = placeholder.mtag

    if 'new_section_name' in call_data:
        if call_data['new_section_name'] == '-None-':
            raise FailPage(message='Section name not set')
        if not section_list:
            raise FailPage(message='No sections found')
        if call_data['new_section_name'] not in section_list:
            raise FailPage(message='Section not found')
        section_name = call_data['new_section_name']
        message = 'New section name set'
    elif 'placeholder_brief' in call_data:
        brief = call_data["placeholder_brief"]
        message = 'New description set'
    elif 'multiplier' in call_data:
        try:
            multiplier = int(call_data["multiplier"])
        except Exception:
            raise FailPage(message='Invalid multiplier, should be an integer of 0 or above')
        if multiplier < 0:
           raise FailPage(message='Invalid multiplier, should be an integer of 0 or above')
        message = 'New multiplier set'
    elif 'placename' in call_data:
        placename = call_data['placename']
        if placename == alias:
            call_data['status'] = 'Alias not changed'
            return
        if not placename:
            raise FailPage("Invalid alias")
        lower_placename = placename.lower()
        if (lower_placename == 'body') or (lower_placename == 'head') or (lower_placename == 'svg')  or (lower_placename == 'show_error'):
            raise FailPage(message="Unable to set alias, the given value is reserved")
        if _AN.search(placename):
            raise FailPage(message="Invalid alias, alphanumeric and underscore only")
        if placename[0] == '_':
            raise FailPage(message="Invalid alias, must not start with an underscore")
        if placename.isdigit():
            raise FailPage(message="Unable to set alias, the value must include some letters")
        alias = placename
        message = 'New alias set'
    elif 'mtag' in call_data:
        mtag = call_data["mtag"]
        message = 'New multiplier container tag set'
    else:
        raise FailPage("A new placeholder value to edit has not been found")

    # call editsection.edit_placeholder from skilift, which returns a new pchange
    try:
        call_data['pchange'] = editsection.edit_placeholder(project, pagenumber, pchange, location, section_name, alias, brief, multiplier, mtag)
    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = message


def retrieve_insert(skicall):
    "Fills in the insert a placeholder page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    if 'page_number' not in call_data:
        raise FailPage("Page to edit not identified")

    pagenumber = call_data['page_number']
    page_info = item_info(project, pagenumber)
    if page_info is None:
        raise FailPage("Page to edit not identified")

    if (page_info.item_type != "TemplatePage") and (page_info.item_type != "SVG"):
        raise FailPage("Page not identified")

    # Fill in header
    page_data[("adminhead","page_head","large_text")] = "Insert Section place holder"

    # get current sections
    section_list = editsection.list_section_names(project)
    if not section_list:
        page_data[("nosection", "show")] = True
        page_data[("descript", "show")] = False
        page_data[("placename","show")] = False
        return

    page_data[('sectionname','option_list')] = section_list[:]
    page_data[('sectionname','selectvalue')] = section_list[0]


def create_insert(skicall):
    "Creates the section placeholder"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    if 'page_number' not in call_data:
        raise FailPage("Page to edit not identified")
    pagenumber = call_data['page_number']
    page_info = item_info(project, pagenumber)
    if page_info is None:
        raise FailPage("Page to edit not identified")

    if (page_info.item_type != "TemplatePage") and (page_info.item_type != "SVG"):
        raise FailPage("Page not identified")

    # create the item

    if 'newsectionname' not in call_data:
        raise FailPage(message = "Missing section name")
    section_name=call_data['newsectionname']

    # get current sections
    section_list = editsection.list_section_names(project)
    if section_name not in section_list:
        raise FailPage(message = "Unknown section name")

    if 'newplacename' not in call_data:
        raise FailPage(message = "Missing section alias")

    alias=call_data['newplacename']

    if _AN.search(alias):
        raise FailPage(message="Invalid alias")

    try:
        call_data['pchange'], call_data['location'] = editsection.new_placeholder(project, pagenumber, call_data['pchange'], call_data['location'], section_name, alias, call_data['newbrief'])
    except ServerError as e:
        raise FailPage(e.message)

    call_data['status'] = "Section placeholder inserted"







"""This module contains basic validation functions

Each Validator is a class, with a _check method

"""


import re
from . import Validator

from ..excepts import ValidateError

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')

# a search for anything none-alphanumeric, not a dot and not an underscore
_AND = re.compile('[^\w\.]')


class AllowedValuesOnly(Validator):
    "Checks the value is in the allowed values only"

    arg_descriptions = {}   # Takes no arguments

    def _check(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        """Item must be one of the allowed values, so this invalidates"""
        return '', False


class NoOperation(Validator):
    "Always passes"

    arg_descriptions = {}   # Takes no arguments

    def _check(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        """Item always validates"""
        return item, True


class AllowedValuesOrEmpty(Validator):
    "Checks the value is in the allowed values or is empty"

    arg_descriptions = {}   # Takes no arguments

    def _check(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        if not item:
            return '', True
        return '', False


class MinLength(Validator):
    "Checks field length is equal or greater than the minimum value given"

    arg_descriptions = {'minlength':1}


    def __setitem__(self, arg_name, value):
        "Sets an argument value"
        Validator.__setitem__(self, arg_name, value)
        try:
            intval = int(value)
        except Exception:
            raise ValidateError(message = 'minlength value must be an integer')
        if intval < 0:
            raise ValidateError(message = 'A minimum value of least 0 is expected')
        self._val_args[arg_name] = intval


    def _check(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        if len(item) < self["minlength"]:
            return '', False
        return item, True


class MaxLength(Validator):
    "Checks field length is equal or less than the maximum value given"

    arg_descriptions = {'maxlength':50}

    def __setitem__(self, arg_name, value):
        "Sets an argument value"
        Validator.__setitem__(self, arg_name, value)
        try:
            intval = int(value)
        except Exception:
            raise ValidateError(message = 'maxlength value must be an integer')
        if intval < 1:
            raise ValidateError(message = 'A maximum value of of at least 1 is expected')
        self._val_args[arg_name] = intval

    def _check(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        if len(item) > self["maxlength"]:
            return '', False
        return item, True


class NotEmpty(Validator):
    "Raises an error if a field is empty"

    arg_descriptions = {}   # Takes no arguments

    def _check(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        if item is False:
            # False is not an empty field
            return item, True
        if item is 0:
            # 0 is not an empty field
            return item, True
        if not item:
            return '', False
        return item, True


class IntMinMax(Validator):
    """Check integer, limit value to between min and max value, raise error if input not an integer string
       but if it is, and is beyond the range, then does not raise an error, just limits the integer size"""

    arg_descriptions = {'maxval':255,
                        'minval':0}


    def __setitem__(self, arg_name, value):
        "Sets an argument value"
        Validator.__setitem__(self, arg_name, value)
        try:
            intval = int(value)
        except Exception:
            raise ValidateError(message = 'values given must be integers')
        if arg_name == 'maxval':
            if intval < 1:
                raise ValidateError(message = 'A maximum value of of at least 1 is expected')
        if arg_name == 'minval':
            if intval < 0:
                raise ValidateError(message = 'A minimum value of least 0 is expected')
        self._val_args[arg_name] = intval

    def _check(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        try:
            i = int(item)
        except Exception:
            return '', False
        if self["maxval"] < self["minval"]:
            return '', False
        if i > self["maxval"]:
            return str(self["maxval"]), False
        if i < self["minval"]:
            return str(self["minval"]), False
        return str(i), True


class AlphaNumUnder(Validator):
    "Checks the value is alphanumric or underscore only"

    arg_descriptions = {}   # Takes no arguments

    def _check(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        "Return item, True if item is alphanumeric or underscore only"
        if  _AN.search(item):
            return '', False
        return item, True


class AlphaNumDotUnder(Validator):
    "Checks the value is alphanumric, dot or underscore only"

    arg_descriptions = {}   # Takes no arguments

    def _check(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        "Return item, True if item is alphanumeric or underscore only"
        if  _AND.search(item):
            return '', False
        return item, True


class Search(Validator):
    """Check item against a regular expresion"""

    arg_descriptions = {'pattern':"[\s\S]"}


    def __setitem__(self, arg_name, value):
        "Sets an argument value"
        Validator.__setitem__(self, arg_name, value)
        if not value:
            raise ValidateError(message = 'A regular expression must be given')
        self._val_args[arg_name] = value

    def _check(self, widgfield, item, environ, lang, form_data, call_data, caller_page_ident):
        "Return item, True if pattern search successfull"
        if re.search(self._val_args['pattern'],item) is None:
            return '', False
        return item, True








"""Contains widgets displaying checkboxes"""

from .. import skiboot, tag, excepts
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict


class CheckBox1(Widget):
    """A div holding a checkbox. Without a form or submit button, typically included within a form"""

    error_location = (0,0,0)

    arg_descriptions = {
                       'checkbox':FieldArg("text", '', valdt=True),
                       'checkbox_class':FieldArg("cssclass", ''),
                       'checkbox_style':FieldArg("cssstyle", ''),
                       'left_label':FieldArg("text", 'Left Label'),
                       'left_class':FieldArg("cssclass", ''),
                       'left_style':FieldArg("cssstyle", ''),
                       'error_class':FieldArg("cssclass", ""),
                       'div_class':FieldArg("cssclass", ""),
                       'div_style':FieldArg("cssstyle", ''),
                       'right_label':FieldArg("text", ''),
                       'right_class':FieldArg("cssclass", ''),
                       'right_style':FieldArg("cssstyle", ''),
                       'checked':FieldArg("boolean", False, jsonset=True)
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        checkbox: The name of the checkbox, with the value returned
        left_label: The text displayed to the left of the checkbox
        left_class: The css class of the label to the left of the checkbox
        checkbox_class: The css class of the checkbox input field
        right_label: The text displayed to the right of the checkbox
        right_class: The css class of the label to the right of the checkbox
        checked: True if checkbox ticked, False otherwise
        error_class: css class applied to the normally hidden error paragraph
        div_class: css class applied to the div after the error paragraph
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        self[1] = tag.Part(tag_name="div")
        self[1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        self[1][1] = tag.ClosedPart(tag_name="input", attribs = {"type":"checkbox"})
        self[1][2] = tag.Part(tag_name="label", hide_if_empty=True)

    def _build(self, page, ident_list, environ, call_data, lang):
        "build the checkbox"
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if self.get_field_value('div_class'):
            self[1].update_attribs({"class":self.get_field_value('div_class')})
        if self.get_field_value('div_style'):
            self[1].update_attribs({"style":self.get_field_value('div_style')})
        if self.get_field_value('left_label'):
            self[1][0][0] = self.get_field_value('left_label')
        if self.get_field_value('left_class'):
            self[1][0].attribs = {"class": self.get_field_value('left_class')}
        if self.get_field_value('left_style'):
            self[1][0].attribs = {"style": self.get_field_value('left_style')}
        if self.get_field_value('checked'):
            self[1][1].update_attribs({"name":self.get_formname('checkbox'), "value":self.get_field_value('checkbox'), "checked":"checked"})
        else:
            self[1][1].update_attribs({"name":self.get_formname('checkbox'), "value":self.get_field_value('checkbox')})
        if self.get_field_value('checkbox_class'):
            self[1][1].update_attribs({"class": self.get_field_value('checkbox_class')})
        if self.get_field_value('checkbox_style'):
            self[1][1].update_attribs({"style": self.get_field_value('checkbox_style')})
        if self.get_field_value('right_label'):
            self[1][2][0] = self.get_field_value('right_label')
        if self.get_field_value('right_class'):
            self[1][2].attribs = {"class": self.get_field_value('right_class')}
        if self.get_field_value('right_style'):
            self[1][2].attribs = {"style": self.get_field_value('right_style')}
        # set an id in the checkbox for the 'label for' tag
        self[1][1].insert_id()
        # set the label 'for' attribute
        self[1][0].update_attribs({'for':self[1][1].get_id()})
        self[1][2].update_attribs({'for':self[1][1].get_id()})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<div>  <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  <div>
  <div>  <!-- with class attribute set to div_class if a class is set -->
    <label> <!-- with class set to left_class and content to left_label -->
    </label>
    <input type='checkbox' /> <!-- checked if checked field is True, and value to the checkbox value -->
                              <!-- and class set to checkbox_class -->
    <label> <!-- with class set to right_class and content to right_label -->
    </label>
  </div>
</div>"""



class CheckBox2(Widget):
    """A span holding a checkbox. Without a form or submit button, typically included within a form"""

    # This class does not display any error messages
    display_errors = False


    arg_descriptions = {
                       'checkbox':FieldArg("text", '', valdt=True),
                       'checkbox_class':FieldArg("cssclass", ''),
                       'checkbox_style':FieldArg("cssstyle", ''),
                       'left_label':FieldArg("text", ''),
                       'left_class':FieldArg("cssclass", ''),
                       'left_style':FieldArg("cssstyle", ''),
                       'right_label':FieldArg("text", ''),
                       'right_class':FieldArg("cssclass", ''),
                       'right_style':FieldArg("cssstyle", ''),
                       'checked':FieldArg("boolean", False, jsonset=True)
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        checkbox: The name of the checkbox, with the value returned
        checkbox_class: The css class of the checkbox input field
        left_label: The text displayed to the left of the checkbox
        left_class: The css class of the label to the left of the checkbox
        right_label: The text displayed to the right of the checkbox
        right_class: The css class of the label to the right of the checkbox
        checked: True if checkbox ticked, False otherwise
        """
        Widget.__init__(self, name=name, tag_name="span", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="label", hide_if_empty=True)
        self[1] = tag.ClosedPart(tag_name="input", attribs = {"type":"checkbox"})
        self[2] = tag.Part(tag_name="label", hide_if_empty=True)


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the checkbox"
        if self.get_field_value('left_label'):
            self[0][0] = self.get_field_value('left_label')
        if self.get_field_value('left_class'):
            self[0].attribs = {"class": self.get_field_value('left_class')}
        if self.get_field_value('left_style'):
            self[0].attribs = {"style": self.get_field_value('left_style')}
        if self.get_field_value('checked'):
            self[1].update_attribs({"name":self.get_formname('checkbox'), "value":self.get_field_value('checkbox'), "checked":"checked"})
        else:
            self[1].update_attribs({"name":self.get_formname('checkbox'), "value":self.get_field_value('checkbox')})
        if self.get_field_value('checkbox_class'):
            self[1].update_attribs({"class": self.get_field_value('checkbox_class')})
        if self.get_field_value('checkbox_style'):
            self[1].update_attribs({"style": self.get_field_value('checkbox_style')})
        if self.get_field_value('right_label'):
            self[2][0] = self.get_field_value('right_label')
        if self.get_field_value('right_class'):
            self[2].attribs = {"class": self.get_field_value('right_class')}
        if self.get_field_value('right_style'):
            self[2].attribs = {"style": self.get_field_value('right_style')}
        # set an id in the checkbox for the 'label for' tag
        self[1].insert_id()
        # set the label 'for' attribute
        self[0].update_attribs({'for':self[1].get_id()})
        self[2].update_attribs({'for':self[1].get_id()})

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<span>  <!-- with widget id and class widget_class -->
  <label> <!-- with class set to left_class and content to left_label -->
  </label>
  <input type='checkbox' /> <!-- checked if checked field is True, and value to the checkbox value -->
                            <!-- and class set to checkbox_class -->
  <label> <!-- with class set to right_class and content to right_label -->
  </label>
</span>"""


class CheckedText(Widget):
    """A div containing a checkbox and text input field. Without a form or submit button, typically included within a form
       The text input field is only enabled when the checkbox is checked
       uses javascript and jquery to enable/disable the text input field"""

    error_location = (0,0,0)

    arg_descriptions = {'checkbox':FieldArg("text", '', valdt=True),
                        'checkbox_class':FieldArg("cssclass", ''),
                        'checked':FieldArg("boolean", False, jsonset=True),
                        'label_text':FieldArg("text", ''),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'error_class':FieldArg("cssclass", ""),
                        'input_text':FieldArg("text", '', valdt=True),
                        'input_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'inputdiv_style':FieldArg("cssstyle", ''),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", '')
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        checkbox: The name of the checkbox, with the value returned, value normally empty, as the checkbox purpose is only
                  to enable the input_text field, however if a value is given, it is returned if checked
        checkbox_class: The css class of the checkbox input field
        label_text: The text displayed to the left of the checkbox
        label_class: The css class of the label to the left of the checkbox
        checked: True if checkbox ticked, False otherwise
        error_class: css class applied to the normally hidden error paragraph
        input_text: The default text in the text input field, field name used as name attribute
        input_class: The css class of the input text field
        inputdiv_class: the class attribute of the tag which contains the label and inputs
        size: The number of characters appearing in the text input area
        maxlength: The maximum number of characters accepted in the text area
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # div containing label and input fields
        self[1] = tag.Part(tag_name='div')
        self[1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        self[1][1] = tag.ClosedPart(tag_name="input", attribs = {"type":"checkbox"})
        self[1][2] = tag.ClosedPart(tag_name="input", attribs ={"type":"text"})

    def _build(self, page, ident_list, environ, call_data, lang):
        "build the checkbox"
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")

        # the div holding label, input fields
        if self.get_field_value('inputdiv_class'):
            self[1].attribs = {"class": self.get_field_value('inputdiv_class')}
        if self.get_field_value('inputdiv_style'):
            self[1].attribs = {"style": self.get_field_value('inputdiv_style')}
        # the label
        if self.get_field_value('label_text'):
            self[1][0][0] = self.get_field_value('label_text')
        if self.get_field_value('label_class'):
            self[1][0].attribs = {"class": self.get_field_value('label_class')}
        if self.get_field_value('label_style'):
            self[1][0].attribs = {"style": self.get_field_value('label_style')}
        if self.get_field_value('checked'):
            self[1][1].update_attribs({"name":self.get_formname('checkbox'), "checked":"checked"})
        else:
            self[1][1].update_attribs({"name":self.get_formname('checkbox')})
            self[1][2].update_attribs({"disabled":"disabled"})
        if self.get_field_value('checkbox'):
            self[1][1].update_attribs({"value":self.get_field_value('checkbox')})
        if self.get_field_value('checkbox_class'):
            self[1][1].update_attribs({"class": self.get_field_value('checkbox_class')})
        self[1][2].update_attribs({"name":self.get_formname('input_text'), "value":self.get_field_value('input_text')})
        if self.get_field_value('size'):
            self[1][2].update_attribs({"size":self.get_field_value('size')})
        if self.get_field_value('maxlength'):
            self[1][2].update_attribs({"maxlength":self.get_field_value('maxlength')})
        if self.get_field_value('input_class'):
            self[1][2].update_attribs({"class":self.get_field_value('input_class')})

        # set an id in the checkbox for the 'label for' tag
        self[1][1].insert_id()
        # set the label 'for' attribute
        self[1][0].update_attribs({'for':self[1][1].get_id()})

    def _build_js(self, page, ident_list, environ, call_data, lang):
        """jscript to enable the textbox"""
        jscript = """  $('#{checkident}').change(function() {{
    $('#{ident} input').last().prop('disabled',!this.checked);
    }});
""".format(checkident=self[1][1].get_id(), ident=self.get_id())
        return jscript

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<div>  <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <div> <!-- class attribute set to inputdiv_class -->
    <label> <!-- with class set to label_class and content to label_text -->
    </label>
    <input type='checkbox' /> <!-- checked if checked field is True, and value to the checkbox value -->
                              <!-- and class set to checkbox_class -->
    <input type=\"text\" /> <!-- class set to input_class and text value set to input_text -->
  </div>
</div>"""


class CheckInputs(Widget):
    """A div holding a checkbox with a container meant to hold further input fields.
       Without a form or submit button, typically included within a form.
       The checkbox controls any text input fields (not other types of input)
       Contained input text fields must be initially set as enabled/disabled to match checkbox
       The contained input fields are only enabled when the checkbox is checked
       uses jquery"""

    _container = ((1,2),)

    error_location = (0,0,0)

    arg_descriptions = {'div_class':FieldArg("cssclass", ''),
                        'checkbox':FieldArg("text", '', valdt=True),
                        'checkbox_class':FieldArg("cssclass", ''),
                        'label_text':FieldArg("text", ''),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'container_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", ""),
                        'checked':FieldArg("boolean", False, jsonset=True)
                       }
    def __init__(self, name=None, brief='', **field_args):
        """
        div_class: The class attribute of the inner div holding input fields.
        checkbox: The name of the checkbox, with the value returned, value normally empty, as the checkbox purpose is only
                  to enable the contained fields, however if a value is given, it is returned if checked
        checkbox_class: The css class of the checkbox input field
        label_text: The text displayed to the left of the checkbox
        label_class: The css class of the label to the left of the checkbox
        container_class: the class attribute of the div holding the container
        checked: True if checkbox ticked, False otherwise
        error_class: css class applied to the normally hidden error paragraph
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        self[1] = tag.Part(tag_name="div")
        self[1][0] = tag.Part(tag_name="label", hide_if_empty=True)
        self[1][1] = tag.ClosedPart(tag_name="input", attribs = {"type":"checkbox"})
        # The location (1,2) is available as a container
        self[1][2] = tag.Part(tag_name="div")
        self[1][2][0] = ""

    def _build(self, page, ident_list, environ, call_data, lang):
        "build the checkbox"
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if self.get_field_value('div_class'):
            self[1].attribs = {"class": self.get_field_value('div_class')}
        if self.get_field_value('label_text'):
            self[1][0][0] = self.get_field_value('label_text')
        if self.get_field_value('label_class'):
            self[1][0].attribs = {"class": self.get_field_value('label_class')}
        if self.get_field_value('label_style'):
            self[1][0].attribs = {"style": self.get_field_value('label_style')}
        # Create the checkbox
        if self.get_field_value('checked'):
            self[1][1].update_attribs({"name":self.get_formname('checkbox'), "checked":"checked"})
        else:
            self[1][1].update_attribs({"name":self.get_formname('checkbox')})
        if self.get_field_value('checkbox'):
            self[1][1].update_attribs({"value":self.get_field_value('checkbox')})
        if self.get_field_value('checkbox_class'):
            self[1][1].update_attribs({"class": self.get_field_value('checkbox_class')})
        # the div holding the container
        if self.get_field_value('container_class'):
            self[1][2].attribs = {"class": self.get_field_value('container_class')}
        # set an id in the checkbox for the 'label for' tag
        self[1][1].insert_id()
        # set the label 'for' attribute
        self[1][0].update_attribs({'for':self[1][1].get_id()})


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """jscript to enable the contained input text fields"""
        jscript = """  $("#{ident} input").first().change(function() {{
    $('#{ident} input[type="text"]').prop('disabled',!this.checked);
    }});
""".format(ident=self.get_id())
        return jscript

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """<div>  <!-- with widget id and class widget_class -->
  <div> <!-- normally hidden div, with class error_class -->
    <p> <!-- Any error text appears here --> </p>
  </div>
  <div>  <!-- with class set to div_class -->
    <label> <!-- with class set to label_class and content to label_text -->
    </label>
    <input type='checkbox' /> <!-- checked if checked field is True, and value to the checkbox value -->
                              <!-- and class set to checkbox_class -->
    <div>
      <!-- container can hold input fields -->
    </div>
  </div>
</div>"""


class SubmitCheckBox1(Widget):
    """Defines a form with a checkbox input field, and four hidden fields"""

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    error_location = (0,0,0)

    arg_descriptions = {'label':FieldArg("text", 'Left Label'),
                        'label_class':FieldArg("cssclass", ''),
                        'label_style':FieldArg("cssstyle", ''),
                        'action_json':FieldArg("url", ''),
                        'action':FieldArg("url", ''),
                        'hidden_field1':FieldArg("text", '', valdt=True),
                        'hidden_field2':FieldArg("text", '', valdt=True),
                        'hidden_field3':FieldArg("text", '', valdt=True),
                        'hidden_field4':FieldArg("text", '', valdt=True),
                        'button_text':FieldArg("text",'Submit'),
                        'button_class':FieldArg("cssclass", ''),
                        'inputdiv_class':FieldArg("cssclass", ''),
                        'inputandbutton_class':FieldArg("cssclass", ''),
                        'error_class':FieldArg("cssclass", ''),
                        'hide':FieldArg("boolean", False, jsonset=True),
                        'checkbox':FieldArg("text", '', valdt=True),
                        'checkbox_class':FieldArg("cssclass", ''),
                        'checked':FieldArg("boolean", False, jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        label: The text displayed to the left of the text input field
        label_class: The css class of the label
        action_json:  if a value set, and client has jscript enabled, this is the page ident, label, url this button links to, expects a json page back
        action: The page ident, label, url this button links to, overridden if action_json is set.
        hidden_field1: A hidden field value, leave blank if unused, name used as the get field name
        hidden_field2: A second hidden field value, leave blank if unused, name used as the get field name
        hidden_field3: A third hidden field value, leave blank if unused, name used as the get field name
        hidden_field4: A fourth hidden field value, leave blank if unused, name used as the get field name
        button_text: The text on the button
        button_class: The class given to the button tag
        inputdiv_class: the class attribute of the div which contains the label, input text and button
        inputandbutton_class: the class attribute of the span which contains the input text and button
        error_class: The class applied to the paragraph containing the error message on error.
        hide: If True, widget is hidden
        checkbox: The name of the checkbox, with the value returned
        checkbox_class: The css class of the checkbox input field
        checked: True if checkbox ticked, False otherwise
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)
        # error div at 0
        self[0] = tag.Part(tag_name="div", attribs={"style":"display:none;"})
        self[0][0] = tag.Part(tag_name="p")
        self[0][0][0] = ''
        # The form
        self[1] = tag.Part(tag_name='form', attribs={"role":"form", "method":"post"})

        # div containing label, input text and button
        self[1][0] = tag.Part(tag_name='div')
        # the label
        self[1][0][0] = tag.Part(tag_name="label", hide_if_empty=True)
        # span containing input text and button
        self[1][0][1] = tag.Part(tag_name='span')
        # the checkbox input field
        self[1][0][1][0] = tag.ClosedPart(tag_name="input", attribs = {"type":"checkbox"})
        # the submit button
        self[1][0][1][1] = tag.Part(tag_name="button", attribs ={"type":"submit"})
        self[1][0][1][1][0] = "Submit"
        self._jsonurl = ''


    def _build(self, page, ident_list, environ, call_data, lang):
        "build the form"
        # Hides widget if no error and hide is True
        self.widget_hide(self.get_field_value("hide"))
        self._jsonurl = skiboot.get_url(self.get_field_value("action_json"), proj_ident=page.proj_ident)
        if self.get_field_value('error_class'):
            self[0].update_attribs({"class":self.get_field_value('error_class')})
        if self.error_status:
            self[0].del_one_attrib("style")
        if not self.get_field_value("action"):
            # setting self._error replaces the entire tag
            self._error = "Warning: No form action"
            return
        actionurl = skiboot.get_url(self.get_field_value("action"),  proj_ident=page.proj_ident)
        if not actionurl:
            # setting self._error replaces the entire tag
            self._error = "Warning: broken link"
            return
        # update the action of the form
        self[1].update_attribs({"action": actionurl})
        # the div holding label, checkbox and button
        if self.get_field_value('inputdiv_class'):
            self[1][0].attribs = {"class": self.get_field_value('inputdiv_class')}
        if self.get_field_value('label_class'):
            self[1][0][0].attribs = {"class": self.get_field_value('label_class')}
        if self.get_field_value('label_style'):
            self[1][0][0].attribs = {"style": self.get_field_value('label_style')}
        if self.get_field_value('label'):
            self[1][0][0][0] = self.get_field_value('label')

        # the span holding input checkbox and button
        if self.get_field_value('inputandbutton_class'):
            self[1][0][1].attribs = {"class": self.get_field_value('inputandbutton_class')}

        # set an id in the input checkbox field for the 'label for' tag
        self[1][0][1][0].insert_id()

        if self.get_field_value('checked'):
            self[1][0][1][0].update_attribs({"name":self.get_formname('checkbox'), "value":self.get_field_value('checkbox'), "checked":"checked"})
        else:
            self[1][0][1][0].update_attribs({"name":self.get_formname('checkbox'), "value":self.get_field_value('checkbox')})
        if self.get_field_value('checkbox_class'):
            self[1][0][1][0].update_attribs({"class": self.get_field_value('checkbox_class')})

        # set the label 'for' attribute
        self[1][0][0].update_attribs({'for':self[1][0][1][0].get_id()})

        # submit button
        if self.get_field_value('button_class'):
            self[1][0][1][1].update_attribs({"class": self.get_field_value('button_class')})
        if self.get_field_value('button_text'):
            self[1][0][1][1][0] = self.get_field_value('button_text')

        # add ident and four hidden fields
        self.add_hiddens(self[1], page)



    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        jscript = """  $("#{ident} form").on("submit input", function(e) {{
    SKIPOLE.widgets["{ident}"].eventfunc(e);
    }});
""".format(ident=self.get_id())
        if self._jsonurl:
            return jscript + self._make_fieldvalues(url=self._jsonurl)
        return jscript


    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div> <!-- with widget id and class widget_class -->
  <div>  <!-- div hidden when no error is displayed, with class set to error_class on error -->
    <p> <!-- error message appears in this paragraph --> </p>
  </div>
  <form role="form" method="post"> <!-- action attribute set to action field -->
    <div> <!-- class attribute set to inputdiv_class -->
      <label> <!-- with class set to label_class and content to label, for set to input checkbox id -->
      </label>
      <span>  <!-- class attribute set to inputandbutton_class -->
          <input type="checkbox" />  <!-- class set to checkbox_class -->
          <button type="submit"> <!-- with class set to button_class -->
            <!-- button_text -->
          </button>
      </span>
    </div>
    <!-- hidden input fields -->                              
  </form>
</div>"""








import os, inspect, re

from skipole import ValidateError, FailPage, ServerError

from skipole import skilift
from skipole.skilift import fromjson, editfolder, editresponder, editpage

from .. import utils, css_styles

# a search for anything none-alphanumeric, not a dot and not an underscore
_AND = re.compile('[^\w\.]')

# a search for anything none-alphanumeric, not a dot and not a underscore and not an hyphen
_ANDH = re.compile('[^\w\.\-]')


def retrieve_add_page(skicall):

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'edited_folder' not in call_data:
        raise FailPage(message = "Folder missing")

    try:
        foldernumber = skilift.get_itemnumber(project, call_data['edited_folder'])
        if foldernumber is None:
            raise FailPage(message="Parent folder not recognised")
        folder_info = skilift.folder_info(project, foldernumber)
        folder_url = skilift.page_path(project, foldernumber)
    except ServerError as e:
        raise FailPage(message=e.message)

    page_data[("adminhead","page_head","large_text")] = "Add page to : %s" % (folder_url,)
    page_data['st1:parent'] = project + "_" + str(foldernumber)

    # rb1: page type checked
    if 'radio_checked' in call_data:
        page_data['rb1:radio_checked'] = call_data['radio_checked']

    page_data[('rb1','radio_values')] = ['page', 'responder', 'svg', 'css', 'json', 'file', 'copy']
    page_data[('rb1','radio_text')] = ['Template page',
                                           'Responder',
                                           'SVG Image page',
                                           'CSS page',
                                           'JSON page',
                                           'File Link page',
                                           'Copy an existing page']

    # st1: new page name
    if 'new_page' in call_data:
        page_data['it4:new_page'] = call_data['new_page']

    # it2: text input for page brief
    if ('page_brief' in call_data) and call_data['page_brief']:
        page_data['it2:page_brief'] = call_data['page_brief']

    # it3: page ident number
    if 'page_ident_number' in call_data:
        page_data['it3:page_ident_number'] = str(call_data['page_ident_number'])
    else:
        page_data['it3:page_ident_number'] = str(skilift.next_ident_number(project))


def _common_page_items(skicall):
    "Returns project, parent foldernumber, new pagenumber, new_name, new_brief"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'edited_folder' not in call_data:
        raise FailPage(message = "Folder missing")
    foldernumber = skilift.get_itemnumber(project, call_data['edited_folder'])
    if foldernumber is None:
        raise FailPage(message="Parent folder not recognised")
    # skilift.folder_info raises a ServerError if this is not a folder
    try:
        folder_info = skilift.folder_info(project, foldernumber)
    except ServerError as e:
        raise FailPage(message=e.message)
    # Get the new page number
    try:
        # pagenumber could be number or "project,number" or "project_number"
        page_ident = call_data['page_ident_number']
        if page_ident.startswith(project):
            if ',' in page_ident:
                page_ident = page_ident.split(',')[-1]
            elif '_' in page_ident:
                page_ident = page_ident.split('_')[-1]
        pagenumber = int(page_ident)
    except Exception:
        raise FailPage(message = "The page number is invalid")
    if pagenumber < 1:
        raise FailPage(message = "The page number is invalid")
    # Get the new page name
    if 'new_page' not in call_data:
        raise FailPage(message = "The name of the new page has not been found")
    new_name = call_data['new_page']
    # check name is alphanumric, dot or underscore only
    if _ANDH.search(new_name):
        raise FailPage(message = "Names must be alphanumric, and may contain dot or underscore or hyphen")
    # Get the new page brief
    if 'page_brief' in call_data:
        new_brief = call_data['page_brief']
    else:
        new_brief = 'New Page'

    return project, foldernumber, pagenumber, new_name, new_brief


def submit_new_template(skicall):
    "Create a new template page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)
    # create page dictionary
    page_dict = _create_templatepagedict(project, new_name, pagenumber, new_brief)
    # create the new page
    try:
        pagenumber = editfolder.make_new_page(project, foldernumber, page_dict)
        fchange = editfolder.folderchange(project, foldernumber)
        pchange = editpage.pagechange(project, pagenumber)
    except ServerError as e:
        raise FailPage(message=e.message)
    # clear and re-populate call_data for edit page
    utils.clear_call_data(call_data)
    call_data['folder_number'] = foldernumber
    call_data['fchange'] = fchange
    call_data['page_number'] = pagenumber
    call_data['pchange'] = pchange
    call_data['status'] = 'Page %s added' % (new_name,)


def _create_templatepagedict(project, name, pagenumber, brief):
    "Returns a dictionary for a new template page"
    backcol = fromjson.get_defaults(project, key="backcol")
    if backcol:
        page_args = {"backcol":backcol}
    else:
        page_args = {}
    page_dict = {"name":name,
                 "ident":pagenumber,
                 "brief":brief,
                 "TemplatePage":page_args,
                 "head":["Part",_head_dict(project, pagenumber)],
                 "body":["Part",_body_dict(project)]
                }
    return page_dict


def _head_dict(project, pagenumber):
    "Returns a dictionary defining the new page head"
    # create the title tag
    title = ["Part", {
                      "tag_name":"title",
                      "brief":"The page title element",
                      "show":True,
                      "hide_if_empty":False,
                      "parts":[ ["Text", "Page " + str(pagenumber)] ]
                     } ]

    # The head dictionary
    head = {
            "tag_name":"head",
            "brief":"The head section of the page",
            "show":True,
            "hide_if_empty":False,
            "parts":[
                        ["ClosedPart", {
                                        "tag_name":"meta",
                                        "brief":"The charset meta declaration",
                                        "show":True,
                                        "attribs":{"charset":"utf-8"}
                                        } ],
                        ["ClosedPart", {
                                        "tag_name":"meta",
                                        "brief":"The viewport meta declaration",
                                        "show":True,
                                        "attribs":{
                                                    "content":"width=device-width, initial-scale=1",
                                                    "name":"viewport"
                                                  }
                                        } ],
                        title,
                    ]
           }

    # append CSS links to the parts list
    css_links = fromjson.get_defaults(project, key="css_links")
    if css_links:
        for label in css_links:
            csslink = ["ClosedPart", {
                                      "tag_name":"link",
                                      "brief":"css link to " + label,
                                      "show":True,
                                      "attribs":{
                                                  "href":"{" + label + "}",
                                                  "rel":"stylesheet",
                                                  "type":"text/css"
                                                 }
                                      } ]
            head["parts"].append(csslink)

    # append Javascript links to the parts list
    js_links = fromjson.get_defaults(project, key="js_links")
    if js_links:
        for label in js_links:
            jslink = ["Part", {
                               "tag_name":"script",
                               "brief":"script link to " + label,
                               "show":True,
                               "hide_if_empty":False,
                               "attribs":{
                                           "src":"{" + label +"}"
                                         },
                               "parts":[]
                              } ]
            head["parts"].append(jslink)

    return head


def _body_dict(project):
    "Returns a dictionary defining the new page body"
    # The body dictionary
    body = {
            "tag_name":"body",
            "brief":"The body section of the page",
            "show":True,
            "hide_if_empty":False,
            "parts":[]
        }
    # if a default body class has been set
    body_class = fromjson.get_defaults(project, key="body_class")
    if body_class:
        body["attribs"] = {"class":body_class}
    return body


def submit_new_svg(skicall):
    """ Creates a new svg page by making a dictionary similar to:

    {
     "name":"page_name",
     "ident":999,
     "brief":"brief description of the page",
     "SVG":{
         "enable_cache":False,
         "width":"100",
         "height":"100"
        }
    }

    And then calling editfolder.make_new_page(project, parent_number, page_dict)

"""

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)
    if 'width' in call_data:
        width = call_data['width']
    else:
        width = '100'
    if 'height' in call_data:
        height = call_data['height']
    else:
        height = '100'
    # create a new page dictionary
    page_dict = {"name":new_name,
                 "ident":pagenumber,
                 "brief":new_brief,
                 "SVG":{
                        "enable_cache":False,
                        "width":width,
                        "height":height
                        }
                }

    # create the new page
    try:
        pagenumber = editfolder.make_new_page(project, foldernumber, page_dict)
        fchange = editfolder.folderchange(project, foldernumber)
        pchange = editpage.pagechange(project, pagenumber)
    except ServerError as e:
        raise FailPage(message=e.message)
    # clear and re-populate call_data for edit page
    utils.clear_call_data(call_data)
    call_data['folder_number'] = foldernumber
    call_data['fchange'] = fchange
    call_data['page_number'] = pagenumber
    call_data['pchange'] = pchange
    call_data['status'] = 'SVG %s added' % (new_name,)


def retrieve_new_svg(skicall):
    "Retrieve data for creating a new SVG page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)

    parent_url = skilift.page_path(project, foldernumber)

    page_data[("adminhead","page_head","large_text")] = "Add SVG image page to : %s" % (parent_url,)

    # information paragraphs
    page_data[('page_name_text','para_text')] = "New page name : " + new_name
    page_data[('page_brief_text','para_text')] = "Description   : " + new_brief

    # information hidden fields
    page_data[('dimensions','hidden_field1')] = str(foldernumber)
    page_data[('dimensions','hidden_field2')] = new_name
    page_data[('dimensions','hidden_field3')] = new_brief
    page_data[('dimensions','hidden_field4')] = str(pagenumber)


def submit_new_css(skicall):
    """ Creates a new css page by making a dictionary similar to:

    {
     "name":"page_name",
     "ident":999,
     "brief":"brief description of the page",
     "CSS":{
         "enable_cache":False,
         "style":{}
        }
    }

    And then calling editfolder.make_new_page(project, parent_number, page_dict)

"""

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)
    # create a new page dictionary
    page_dict = {"name":new_name,
                 "ident":pagenumber,
                 "brief":new_brief,
                 "CSS":{
                        "enable_cache":False
                       }
                }
    # create the new page
    try:
        pagenumber = editfolder.make_new_page(project, foldernumber, page_dict)
        fchange = editfolder.folderchange(project, foldernumber)
        pchange = editpage.pagechange(project, pagenumber)
    except ServerError as e:
        raise FailPage(message=e.message)
    # clear and re-populate call_data for edit page
    utils.clear_call_data(call_data)
    call_data['folder_number'] = foldernumber
    call_data['fchange'] = fchange
    call_data['page_number'] = pagenumber
    call_data['pchange'] = pchange
    call_data['status'] = 'CSS page %s added' % (new_name,)


def submit_new_json(skicall):
    """Create a new json page by making a dictionary similar to:

    {
     "name":"page_name",
     "ident":999,
     "brief":"brief description of the page",
     "JSON":{
            "enable_cache":False,
            "content":{}
        }
    }

    And then calling editfolder.make_new_page(project, parent_number, page_dict)
"""

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)
    # create a new page dictionary
    page_dict = {"name":new_name,
                 "ident":pagenumber,
                 "brief":new_brief,
                 "JSON":{
                        "enable_cache":False
                       }
                }
    # create the new page
    try:
        pagenumber = editfolder.make_new_page(project, foldernumber, page_dict)
        fchange = editfolder.folderchange(project, foldernumber)
    except ServerError as e:
        raise FailPage(message=e.message)
    # clear and re-populate call_data
    utils.clear_call_data(call_data)
    call_data['folder_number'] = foldernumber
    call_data['fchange'] = fchange
    # Currently, after adding a json page, go back to edit folder
    call_data['status'] = 'JSON page %s added' % (new_name,)


def submit_new_file(skicall):
    """Create a new file page by making a dictionary similar to:

    {
     "name":"page_name",
     "ident":999,
     "brief":"brief description of the page",
     "FilePage":{
         "enable_cache":False,
         "filepath": "project/static/filename"
        }
    }

    And then calling editfolder.make_new_page(project, parent_number, page_dict)
"""

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)
    # get the submitted filepath this page links to
    filepath = call_data['filepath']
    # create a new page dictionary
    page_dict = {"name":new_name,
                 "ident":pagenumber,
                 "brief":new_brief,
                 "FilePage":{
                        "enable_cache":False,
                        "filepath": filepath
                       }
                }
    # create the new page
    try:
        pagenumber = editfolder.make_new_page(project, foldernumber, page_dict)
        fchange = editfolder.folderchange(project, foldernumber)
        pchange = editpage.pagechange(project, pagenumber)
    except ServerError as e:
        raise FailPage(message=e.message)
    # clear and re-populate call_data for edit page
    utils.clear_call_data(call_data)
    call_data['folder_number'] = foldernumber
    call_data['fchange'] = fchange
    call_data['page_number'] = pagenumber
    call_data['pchange'] = pchange
    call_data['status'] = 'File link page %s added' % (new_name,)


def retrieve_new_file(skicall):
    "Gets data for a create filepage"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)

    parent_url = skilift.page_path(project, foldernumber)

    page_data[("adminhead","page_head","large_text")] = "Add File Link page to : %s" % (parent_url,)

    # information paragraphs
    page_data['page_name_text:para_text'] = "New page name : " + new_name
    page_data['page_brief_text:para_text'] = "Description   : " + new_brief

    # information hidden fields
    page_data['setfilepath','pagename'] = new_name
    page_data['setfilepath','pagebrief'] = new_brief
    page_data['setfilepath','pageident'] = str(pagenumber)

    # top descriptive text
    page_data['top_text:para_text'] = """The filepath set here should be relative to the projectfiles folder, so for a file
called myfile in the static directory, the path should be %s""" % (os.path.join(project, 'static', "myfile"),)

    # the submit filepath text input
    page_data['setfilepath:input_text'] = os.path.join(project, 'static', new_name)
    page_data['setfilepath:parent'] = str(foldernumber)


def submit_new_responder(skicall):
    """Create a new responder page by making a dictionary similar to:

    {
     "name":"page_name",
     "ident":999,
     "brief":"brief description of the page",
     "RespondPage":{
         "class":"ResponderClass",
         "original_args": {},
         "original_fields": {}
        }
    }

    And then calling editfolder.make_new_page(project, parent_number, page_dict)
"""

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)
    # responder class name
    responder_class = call_data['responder_class']
    # create a new page dictionary
    page_dict = {"name":new_name,
                 "ident":pagenumber,
                 "brief":new_brief,
                 "RespondPage":{
                    "class":responder_class,
                    "original_args": {},
                    "original_fields": {}
                    }
                }
    # create the new page
    try:
        pagenumber = editfolder.make_new_page(project, foldernumber, page_dict)
        fchange = editfolder.folderchange(project, foldernumber)
        pchange = editpage.pagechange(project, pagenumber)
    except ServerError as e:
        raise FailPage(message=e.message)
    # clear and re-populate call_data for edit page
    utils.clear_call_data(call_data)
    call_data['folder_number'] = foldernumber
    call_data['fchange'] = fchange
    call_data['page_number'] = pagenumber
    call_data['pchange'] = pchange
    call_data['status'] = 'Responder %s - type %s added' % (new_name, responder_class)



def retrieve_new_responder(skicall):
    "Gets data for a create respondpage"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)

    parent_url = skilift.page_path(project, foldernumber)

    page_data[("adminhead","page_head","large_text")] = "Add a responder page to : %s" % (parent_url,)

    # information paragraphs
    page_data['page_name_text:para_text'] = "New page name : " + new_name
    page_data['page_brief_text:para_text'] = "Description   : " + new_brief

    # get a list of lists, each inner being [responder name, module name] from skilift.editresponder
    responderlist = editresponder.list_responders()

    # Create a list of 1) the responder class name, being the text to place on a button
    #                  2) the textblock reference describing the responder

    page_data['responderlinks','buttons'] =[]
    for responder in responderlist:
        rbutton = [responder[0],   "responders." + responder[1] + "." + responder[0]]
        page_data['responderlinks','buttons'].append(rbutton)

    # the hidden fields
    page_data['responderlinks:hidden_field1'] = str(foldernumber)
    page_data['responderlinks:hidden_field2'] = new_name
    page_data['responderlinks:hidden_field3'] = new_brief
    page_data['responderlinks:hidden_field4'] = str(pagenumber)


def submit_copy_page(skicall):
    "Copy a page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, new_page_number, new_name, new_brief = _common_page_items(skicall)
    # Get the page to be copied
    if 'copyident' not in call_data:
        raise FailPage(message = "The ident of the page to be copied has not been found", widget='copyident')
    pagenumber = call_data['copyident']
    try:
        fchange = editfolder.copy_page(project, pagenumber, foldernumber, new_page_number, new_name, new_brief)
    except ServerError as e:
        raise FailPage(e.message)
    edited_folder = call_data['edited_folder']
    utils.clear_call_data(call_data)
    call_data['fchange'] = fchange
    call_data['edited_folder'] = edited_folder
    call_data['folder_number'] = foldernumber
    call_data['status'] = 'Page %s added' % (new_name,)


def retrieve_new_copypage(skicall):
    "Gets data for a create a page copy"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)
    parent_url = skilift.page_path(project, foldernumber)

    page_data[("adminhead","page_head","large_text")] = "Add a page copy to : %s" % (parent_url,)

    # information paragraphs
    page_data['page_name_text:para_text'] = "New page name : " + new_name
    page_data['page_brief_text:para_text'] = "Description   : " + new_brief

    # information hidden fields
    page_data['copyident','pagename'] = new_name
    page_data['copyident','pagebrief'] = new_brief
    page_data['copyident','pageident'] = str(pagenumber)
    page_data['copyident','parent'] = str(foldernumber)

    page_data['upload','pagename'] = new_name
    page_data['upload','pagebrief'] = new_brief
    page_data['upload','pageident'] = str(pagenumber)
    page_data['upload','parent'] = str(foldernumber)

    # top descriptive text
    page_data['top_text:para_text'] = """To copy a page, either the ident number of an existing page to be copied is required, alternatively - a previously downloaded page definition file can be uploaded."""

    # the submit ident text input
    if 'copyident' in call_data:
        page_data['copyident','input_text'] = call_data['copyident']


def submit_upload_page(skicall):
    "Copy a page from uploaded file"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # first get submitted data for the new page
    project, foldernumber, pagenumber, new_name, new_brief = _common_page_items(skicall)
    # get uploaded file contents
    if "upload" not in call_data:
        raise FailPage("upload missing from call data")
    file_contents = call_data["upload"]
    json_string = file_contents.decode(encoding='utf-8')
    # create the page
    try:
        fromjson.create_page(project, foldernumber, pagenumber, new_name, new_brief, json_string)
        fchange = editfolder.folderchange(project, foldernumber)
    except ServerError as e:
        raise FailPage(message = e.message)
    edited_folder = call_data['edited_folder']
    utils.clear_call_data(call_data)
    call_data['edited_folder'] = edited_folder
    call_data['folder_number'] = foldernumber
    call_data['fchange'] = fchange
    call_data['status'] = 'Page %s added' % (new_name,)








"Functions implementing respond page editing"

import collections


from skipole import skilift
from skipole.skilift import editresponder

from skipole import ValidateError, FailPage, ServerError, GoTo

from .. import utils

def _ident_to_str(ident):
    "Returns string ident or label"
    if ident is None:
        return ''
    if isinstance(ident, str):
        return ident
    # ident must be a list or tuple of (project,number)
    if len(ident) != 2:
        raise FailPage("Invalid ident")
    return ident[0] + "," + str(ident[1])


def _field_to_string(wfield):
    "Returns two forms of a widgfield, or if a string, then just the string twice"
    if isinstance(wfield, str):
        return wfield, wfield
    # a widgfield has four elements, reduce it to the non empty elements
    shortwfield = [ w for w in wfield if w ]
    if len(shortwfield) == 1:
        return shortwfield[0], shortwfield[0]
    wf1 = ",".join(shortwfield)
    if len(shortwfield) == 2:
        wf2 = shortwfield[0] + ":" + shortwfield[1]
    else:
        wf2 = shortwfield[0] + "-" + shortwfield[1] + ":" + shortwfield[2]
    return wf1, wf2


def _t_ref(r_info, item):
    "Returns a TextBlock ref for the given item"
    return ".".join(["responders", r_info.module_name, r_info.responder, item])



def skicall_help(skicall):
    "Retrieves help text for the skicall object"
    text = skicall.textblock("aboutcode.skicall")
    if not text:
        text = "No help text for aboutcode.skicall has been found"
    skicall.page_data[("adminhead","show_help","para_text")] = "\n" + text
    skicall.page_data[("adminhead","show_help","hide")] = False


def use_submit_list_help(skicall):
    "Retrieves help text for the use_submit_list decorator"
    text = skicall.textblock("aboutcode.usesubmitlist")
    if not text:
        text = "No help text for aboutcode.use_submit_list has been found"
    skicall.page_data[("adminhead","show_help","para_text")] = "\n" + text
    skicall.page_data[("adminhead","show_help","hide")] = False


def fail_page_help(skicall):
    "Retrieves help text for the fail page ident"
    text = skicall.textblock("responders.fail_page")
    if not text:
        text = "No help text for responders.fail_page has been found"
    skicall.page_data[("adminhead","show_help","para_text")] = "\n" + text
    skicall.page_data[("adminhead","show_help","hide")] = False


def submit_data_help(skicall):
    "Retrieves help text for the submit_data function"
    call_data = skicall.call_data
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message = "page missing")
    try:
        project = call_data['editedprojname']
        # get a ResponderInfo named tuple with information about the responder
        r_info = editresponder.responder_info(project, pagenumber, call_data['pchange'])
    except ServerError as e:
        raise FailPage(message=e.message)
    sdtextref = _t_ref(r_info, 'submit_data')
    text = skicall.textblock(sdtextref)
    if not text:
        text = "No help text for %s has been found" % sdtextref
    skicall.page_data[("adminhead","show_help","para_text")] = "\n" + text
    skicall.page_data[("adminhead","show_help","hide")] = False


def submit_dict_help(skicall):
    "Retrieves help text for the responder submit_dict"
    call_data = skicall.call_data
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message = "page missing")
    try:
        project = call_data['editedprojname']
        # get a ResponderInfo named tuple with information about the responder
        r_info = editresponder.responder_info(project, pagenumber, call_data['pchange'])
    except ServerError as e:
        raise FailPage(message=e.message)
    sdtextref = _t_ref(r_info, 'submit_dict')
    text = skicall.textblock(sdtextref)
    if not text:
        text = "No help text for %s has been found" % sdtextref
    skicall.page_data[("adminhead","show_help","para_text")] = "\n" + text
    skicall.page_data[("adminhead","show_help","hide")] = False


def call_data_help(skicall):
    "Retrieves help text for the responder call_data"
    call_data = skicall.call_data
    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
    else:
        raise FailPage(message = "page missing")
    try:
        project = call_data['editedprojname']
        # get a ResponderInfo named tuple with information about the responder
        r_info = editresponder.responder_info(project, pagenumber, call_data['pchange'])
    except ServerError as e:
        raise FailPage(message=e.message)
    cdtextref = _t_ref(r_info, 'call_data')
    text = skicall.textblock(cdtextref)
    if not text:
        text = "No help text for %s has been found" % cdtextref
    skicall.page_data[("adminhead","show_help","para_text")] = "\n" + text
    skicall.page_data[("adminhead","show_help","hide")] = False


def retrieve_edit_respondpage(skicall):
    "Retrieves widget data for the edit respond page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # clears any session data, keeping page_number, pchange and any status message
    utils.clear_call_data(call_data, keep=["page_number", "pchange", "status"])

    if 'page_number' in call_data:
        pagenumber = call_data['page_number']
        str_pagenumber = str(pagenumber)
    else:
        raise FailPage(message = "page missing")

    try:
        project = call_data['editedprojname']
        pageinfo = skilift.page_info(project, pagenumber)

        if pageinfo.item_type != 'RespondPage':
            raise FailPage(message = "Invalid page")

        call_data['pchange'] = pageinfo.change

        # fills in the data for editing page name, brief, parent, etc., 
        page_data[("adminhead","page_head","large_text")] = pageinfo.name
        page_data[('page_edit','p_ident','page_ident')] = (project,str_pagenumber)
        page_data[('page_edit','p_name','page_ident')] = (project,str_pagenumber)
        page_data[('page_edit','p_description','page_ident')] = (project,str_pagenumber)
        page_data[('page_edit','p_rename','input_text')] = pageinfo.name
        page_data[('page_edit','p_parent','input_text')] = "%s,%s" % (project, pageinfo.parentfolder_number)
        page_data[('page_edit','p_brief','input_text')] = pageinfo.brief

        # get a ResponderInfo named tuple with information about the responder
        r_info = editresponder.responder_info(project, pagenumber, call_data['pchange'])
    except ServerError as e:
        raise FailPage(message=e.message)

    page_data['respondertype:para_text'] = "This page is a responder of type: %s." % (r_info.responder,)
    page_data['responderdescription:textblock_ref'] = ".".join(["responders",r_info.module_name, r_info.responder])

    if r_info.widgfield_required:
        if r_info.widgfield:
            page_data['widgfield:input_text'] = r_info.widgfield
    else:
        page_data['widgfield:show'] = False

    if r_info.alternate_ident_required:
        page_data['alternate','input_text'] = _ident_to_str(r_info.alternate_ident)
        page_data['alternate_ident_description','textblock_ref'] = _t_ref(r_info, 'alternate_ident')
    else:
        page_data['alternate','show'] = False
        page_data['alternate_ident_description','show'] = False

    if r_info.target_ident_required:
        page_data['target:input_text'] = _ident_to_str(r_info.target_ident)
        page_data['target_ident_description','textblock_ref'] = _t_ref(r_info, 'target_ident')
    else:
        page_data['target:show'] = False
        page_data['target_ident_description','show'] = False

    if r_info.allowed_callers_required:
        page_data['allowed_callers_description:textblock_ref'] = _t_ref(r_info, 'allowed_callers')
        if r_info.allowed_callers:
            contents = []
            for ident in r_info.allowed_callers:
                ident_row = [_ident_to_str(ident), _ident_to_str(ident).replace(",","_")]
                contents.append(ident_row)
            page_data['allowed_callers_list:contents'] = contents
        else:
            page_data['allowed_callers_list:show'] = False
        page_data['add_allowed_caller:input_text'] = ''
    else:
        page_data['allowed_callers_description:show'] = False
        page_data['allowed_callers_list:show'] = False
        page_data['add_allowed_caller:show'] = False

    # validate option
    if r_info.validate_option_available:
        page_data[('val_option_desc', 'textblock_ref')] =  _t_ref(r_info, 'validate_option')
        if r_info.validate_option:
            page_data['set_val_option','button_text'] = "Disable Validation"
            page_data['val_status','para_text'] = "Validate received field values : Enabled"
            page_data['validate_fail', 'input_text'] = _ident_to_str(r_info.validate_fail_ident)
            page_data['validate_fail', 'hide'] = False
        else:
            page_data['set_val_option','button_text'] = "Enable Validation"
            page_data['val_status','para_text'] = "Validate received field values : Disabled"
            page_data['validate_fail', 'hide'] = True
    else:
        page_data['val_option_desc','show'] = False
        page_data['set_val_option','show'] = False
        page_data['val_status','show'] = False
        page_data['validate_fail', 'show'] = False

    
    # submit option

    if r_info.submit_option_available:
        page_data['submit_option_desc','textblock_ref'] = _t_ref(r_info, 'submit_option')
        if r_info.submit_option:
            page_data['set_submit_option','button_text'] = 'Disable submit_data'
            page_data['submit_status','para_text'] = "Call submit_data : Enabled"
        else:
            page_data['set_submit_option','button_text'] = 'Enable submit_data'
            page_data['submit_status','para_text'] = "Call submit_data : Disabled"
    else:
        page_data['submit_option_desc','show'] = False
        page_data['set_submit_option','show'] = False
        page_data['submit_status','show'] = False

    if r_info.submit_required or r_info.submit_option:
        page_data['submit_list_description','textblock_ref'] = 'responders.about_submit_list'

        if r_info.submit_list:
            contents = []
            for index, s in enumerate(r_info.submit_list):
                s_row = [s, str(index)]
                contents.append(s_row)
            page_data['submit_list','contents'] = contents
        else:
            page_data['submit_list','show'] = False
        page_data['submit_string','input_text'] = ''
        page_data['fail_page_ident:input_text'] = _ident_to_str(r_info.fail_ident)
    else:
        page_data['submit_list_description','show'] = False
        page_data['submit_list','show'] = False
        page_data['submit_string','show'] = False
        page_data['submit_info','show'] = False
        page_data['fail_page_ident','show'] = False

    # final paragraph
    page_data['final_paragraph:textblock_ref'] = _t_ref(r_info, 'final_paragraph')

    # field options
    f_options = r_info.field_options
    if not f_options['fields']:
        # no fields so no further data to input
        return page_data

    # show the fields description
    page_data['fields_description:show'] = True
    page_data['fields_description:textblock_ref'] = _t_ref(r_info, 'fields')


    if f_options['field_values'] and ( not f_options['single_field'] ):
        page_data['field_values_list:show'] = True
        page_data['add_field_value:show'] = True
        page_data['add_field_value_para:show'] = True
        # populate field_values_list
        contents = []
        field_vals = r_info.field_values_list
        for field, value in field_vals:
            f1,f2 = _field_to_string(field)
            v1,v2 = _field_to_string(value)
            if not v1:
                v1 = "' '"
            row = [f1, v1, f2]
            contents.append(row)
        if contents:
            contents.sort()
            page_data['field_values_list:contents'] = contents
        else:
            page_data['field_values_list:show'] = False
        # populate add_field_value
        if f_options['widgfields']:
            if f_options['widgfield_values']:
                 page_data['add_field_value_para:para_text'] = "Add widgfields in both the field area and in the value area"
            else:
                page_data['add_field_value_para:para_text'] = "Add widgfield and value"
        else:
            page_data['add_field_value_para:para_text'] = "Add items"
            page_data['add_field_value','label'] = "Items:"
        if f_options['empty_values_allowed']:
           page_data['add_field_value_para:para_text'] += ", empty values are allowed:"
        else:
           page_data['add_field_value_para:para_text'] += ", empty values are not allowed:"
        return
       

    if (not f_options['field_values']) and (not f_options['single_field']):
        page_data['field_list','show'] = True
        page_data['add_field','show'] = True
        # populate field_list
        contents = []
        field_vals = r_info.field_list
        for field in field_vals:
            f1,f2 = _field_to_string(field)
            row = [f1, f2]
            contents.append(row)
        if contents:
            contents.sort()
            page_data['field_list','contents'] = contents
        else:
            page_data['field_list','show'] = False
        # populate add_field
        if f_options['widgfields']:
             page_data['add_field','label'] = "Add a widgfield:"
        else:
            page_data['add_field,','label'] = "Add a string:"

    if not f_options['single_field']:
        return

    # all remaining are if a single field is required

    if not f_options['field_values']:
        # a single value is a submit text input field
        page_data[('single_field','show')] = True
        if not r_info.single_field:
            page_data[('single_field','input_text')] = ''
        else:
            page_data[('single_field','input_text')] = r_info.single_field
        return


    # remaining is for a single field and value - still to do
    # as there is currently no responder which takes a single field and value
    #????????????????????????????????????????????????????????


def submit_widgfield(skicall):
    "Sets widgfield"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'widget' in call_data:
        raise FailPage(message="No widgfield given", widget="widgfield_error")
    if not call_data['widget']:
        raise FailPage(message="No widgfield given", widget="widgfield_error")
    # Set the page widgfield
    try:
        call_data['pchange'] = editresponder.set_widgfield(project, pagenumber, pchange, call_data['widget'])
    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = 'WidgField set'


def submit_alternate_ident(skicall):
    "Sets the alternate page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'alternate_ident' in call_data:
        raise FailPage(message="No alternate page label given", widget="alternate")
    # Set the page alternate_ident
    try:
        call_data['pchange'] = editresponder.set_alternate_ident(project, pagenumber, pchange, call_data['alternate_ident'])
    except ServerError as e:
        raise FailPage(e.message)
    page_data['alternate','set_input_accepted'] = True
    call_data['status'] = 'Page set'



def submit_target_ident(skicall):
    "Sets the target ident"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'target_ident' in call_data:
        raise FailPage(message="No target ident given", widget="target")
    # Set the page target_ident
    try:
        call_data['pchange'] = editresponder.set_target_ident(project, pagenumber, pchange, call_data['target_ident'])
    except ServerError as e:
        raise FailPage(e.message)
    page_data['target','set_input_accepted'] = True
    call_data['status'] = 'Target Ident set'


def submit_validate_fail_ident(skicall):
    "Sets the validate fail ident"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'validate_fail_ident' in call_data:
        raise FailPage(message="No validate fail ident given", widget="validate_fail")
    # Set the page validate_fail_ident
    try:
        call_data['pchange'] = editresponder.set_validate_fail_ident(project, pagenumber, pchange, call_data['validate_fail_ident'])
    except ServerError as e:
        raise FailPage(e.message)
    page_data['validate_fail','set_input_accepted'] = True
    call_data['status'] = 'Validate Fail Ident set'


def submit_fail_ident(skicall):
    "Sets the fail ident"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'fail_ident' in call_data:
        raise FailPage(message="No fail ident given", widget="fail_ident_error")
    # Set the page fail_ident
    try:
        call_data['pchange'] = editresponder.set_fail_ident(project, pagenumber, pchange, call_data['fail_ident'])
    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = 'Fail Ident set'


def add_allowed_caller(skicall):
    "Adds a new allowed caller"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'allowed_caller' in call_data:
        raise FailPage(message="No allowed caller given", widget="allowed_callers_error")
    # Set the page allowed caller
    try:
        call_data['pchange'] = editresponder.add_allowed_caller(project, pagenumber, pchange, call_data['allowed_caller'])
    except ServerError as e:
        raise FailPage(e.message)


def delete_allowed_caller(skicall):
    "Deletes an allowed caller"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'delete_allowed_caller' in call_data:
        raise FailPage(message="No allowed caller given", widget="allowed_callers_error")
    # Delete the page allowed caller
    try:
        call_data['pchange'] = editresponder.delete_allowed_caller(project, pagenumber, pchange, call_data['delete_allowed_caller'])
    except ServerError as e:
        raise FailPage(e.message)


def remove_field(skicall):
    "Deletes a field"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'remove_field' in call_data:
        raise FailPage(message="No field to remove given", widget="fields_error")
    # Delete the page field
    try:
        call_data['pchange'] = editresponder.remove_field(project, pagenumber, pchange, call_data['remove_field'])
    except ServerError as e:
        raise FailPage(e.message)


def add_field_value(skicall):
    "Adds a field and value"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']

    if not 'field' in call_data:
        raise FailPage(message="No field given", widget="fields_error")
    if not call_data['field']:
        raise FailPage(message="No field given", widget="fields_error")
    if not 'value' in call_data:
        raise FailPage(message="No value given", widget="fields_error")

    # if value is empty ensure empty values allowed
    if not call_data['value']:
        # get a ResponderInfo named tuple with information about the responder
        try:
            r_info = editresponder.responder_info(project, pagenumber, pchange)
        except ServerError as e:
            raise FailPage(message=e.message)
        # field options
        f_options = r_info.field_options
        if not f_options['fields']:
            raise FailPage(message="Invalid submission, this responder does not have fields", widget="fields_error")
        if not f_options['empty_values_allowed']:
            page_data['add_field_value', 'input_text1'] = call_data['field']
            page_data['add_field_value', 'set_input_accepted1'] = True
            page_data['add_field_value', 'set_input_errored2'] = True
            raise FailPage(message="Invalid submission, empty field values are not allowed", widget="fields_error")
    # Add the field and value
    try:
        call_data['pchange'] = editresponder.add_field_value(project, pagenumber, pchange, call_data['field'], call_data['value'])
    except ServerError as e:
        raise FailPage(e.message)


def add_field(skicall):
    "Adds a field"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'field' in call_data:
        raise FailPage(message="No field given", widget="fields_error")
    if not call_data['field']:
        raise FailPage(message="No field given", widget="fields_error")
    # Add the field
    try:
        call_data['pchange'] = editresponder.add_field(project, pagenumber, pchange, call_data['field'])
    except ServerError as e:
        raise FailPage(e.message)


def set_single_field(skicall):
    "Sets the field in a responder, which require s single field only"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not ('single_field', 'input_text') in call_data:
        raise FailPage(message="No field given", widget="fields_error")
    field = call_data[('single_field', 'input_text')]
    if not field:
        raise FailPage(message="No field given", widget="fields_error")
    # Add the field
    try:
        call_data['pchange'] = editresponder.set_single_field(project, pagenumber, pchange, field)
    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = 'Fields set'



def delete_submit_list_string(skicall):
    "deletes an indexed string from the submit_list"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'delete_submit_list_string_index' in call_data:
        raise FailPage(message="No submit_list string given", widget="submit_list_error")
    try:
        # get the submit list
        submit_list = editresponder.get_submit_list(project, pagenumber, pchange)
        idx = int(call_data['delete_submit_list_string_index'])
        del submit_list[idx]
        call_data['pchange'] = editresponder.set_submit_list(project, pagenumber, pchange, submit_list)
    except ServerError as e:
        raise FailPage(e.message)


def add_submit_list_string(skicall):
    "Adds a new submit_list string"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    if not 'submit_list_string' in call_data:
        raise FailPage(message="No submit_list string given", widget="submit_list_error")
    try:
        # get the submit list
        submit_list = editresponder.get_submit_list(project, pagenumber, pchange)
        submit_list.append(call_data['submit_list_string'])
        call_data['pchange'] = editresponder.set_submit_list(project, pagenumber, pchange, submit_list)
    except ServerError as e:
        raise FailPage(e.message)


def set_validate_option(skicall):
    "Enable or disable the validate option"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    try:
        call_data['pchange'], validate_option = editresponder.toggle_validate_option(project, pagenumber, pchange)
    except ServerError as e:
        raise FailPage(e.message)        
    if validate_option:
        page_data['set_val_option','button_text'] = "Disable Validation"
        page_data['val_status','para_text'] = "Validate received field values : Enabled"
        page_data['validate_fail', 'hide'] = False
    else:
        page_data['set_val_option','button_text'] = "Enable Validation"
        page_data['val_status','para_text'] = "Validate received field values : Disabled"
        page_data['validate_fail', 'hide'] = True
    call_data['status'] = 'Validator changed'


def set_submit_option(skicall):
    "Enable or disable the submit option"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = call_data['page_number']
    pchange = call_data['pchange']
    try:
        call_data['pchange'], submit_option = editresponder.toggle_submit_option(project, pagenumber, pchange)
    except ServerError as e:
        raise FailPage(e.message)  
    call_data['status'] = 'Submit option changed'


def map(skicall):
    "Creates the responder map"
    pagenumber = skicall.call_data['page_number']
    project = skicall.call_data['editedprojname']
    page_data = skicall.page_data
    page_data['responder', 'responderid', 'text'] = "Ident: " + str(pagenumber)

    # insert font text style
    page_data['textstyle', 'text'] = """
  <style>
    /* <![CDATA[ */
    text {
      fill: black;
      font-family: Arial, Helvetica, sans-serif;
    }
    .bigtext {
      font-size: 20px;
    }
    /* ]]> */
  </style>
    """


    map_height = 1600

    # get information about the responder
    pageinfo = skilift.page_info(project, pagenumber)
    r_info = editresponder.responder_info(project, pagenumber)
    i_info = skilift.item_info(project, pagenumber)
    label_list = i_info.label_list

    # list of all responders
    responder_list = editresponder.all_responders(project)

    # Find all responders which call this responder

    callers = [[0,0, "Responders in this project with %s as Target:" % pagenumber]]
    callers2 = []
    n = 40
    for responder_id in responder_list:
        responder_info = editresponder.responder_info(project, responder_id)
        target = responder_info.target_ident
        if target:
            if isinstance(target, str) and (target in label_list):
                moreinfo = skilift.page_info(project, responder_id)
                if n<=300:
                    callers.append([0,n,str(responder_id) + " " + moreinfo.brief])
                else:
                    callers2.append([0,n-280,str(responder_id) + " " + moreinfo.brief])
                n += 20
            elif isinstance(target, tuple) and (len(target) == 2) and (project == target[0]) and (pagenumber == target[1]):
                moreinfo = skilift.page_info(project, responder_id)
                if n<=300:
                    callers.append([0,n,str(responder_id) + " " + moreinfo.brief])
                else:
                    callers2.append([0,n-280,str(responder_id) + " " + moreinfo.brief])
                n += 20
    if n == 40:
        page_data['callers','show'] = False
        page_data['callers2','show'] = False
    elif not callers2:
        page_data['callers', 'callers', 'lines'] = callers
        page_data['callers2','show'] = False
    else:
        page_data['callers', 'callers', 'lines'] = callers
        page_data['callers2', 'callers', 'lines'] = callers2


    # Find all responders which call this responder on failure

    fails = [[0,0, "Responders in this project with %s as Fail Page:" % pagenumber]]
    n = 40
    count = 0
    for responder_id in responder_list:
        responder_info = editresponder.responder_info(project, responder_id)
        failident = responder_info.fail_ident
        if failident:
            if n > 300:
                # do not display more than 14 responders, but continue to count remaining ones
                count += 1
                continue
            if isinstance(failident, str) and (failident in label_list):
                moreinfo = skilift.page_info(project, responder_id)
                fails.append([0,n,str(responder_id) + " " + moreinfo.brief])
                n += 20
            elif isinstance(failident, tuple) and (len(failident) == 2) and (project == failident[0]) and (pagenumber == failident[1]):
                moreinfo = skilift.page_info(project, responder_id)
                fails.append([0,n,str(responder_id) + " " + moreinfo.brief])
                n += 20
    if count:
        fails.append([0, 320, "Plus %s more responders." % (count,)]) 
    if n == 40:
        page_data['fails','show'] = False
    else:
        page_data['fails', 'callers', 'lines'] = fails


    # Find allowed callers to this responder

    allowed_list = r_info.allowed_callers
    if allowed_list:
        page_data['allowed','show'] = True
        allowed = [[0,0, "Allowed callers to %s:" % pagenumber], [0,20, "(Calling page must provide ident information)"]]
        n = 40
        for allowedid in allowed_list:
            allowedident = allowedid
            if isinstance(allowedident, str):
                allowedident = skilift.ident_from_label(project, allowedident)
            if allowedident is None:
                allowed.append([0,n,"UNKNOWN page: " + allowedid])
                n += 20
            elif isinstance(allowedident, str):
                allowed.append([0,n,"INVALID ident: " + allowedident])
                n += 20
            elif isinstance(allowedident, tuple) and (len(allowedident) == 2):
                allowedinfo = skilift.page_info(*allowedident)
                if allowedident[0] == project:
                    allowed.append([0,n,str(allowedident[1]) + ": " + allowedinfo.brief])
                else:
                    allowed.append([0,n,allowedident[0] + ", " + str(allowedident[1]) + ": " + allowedinfo.brief])
                n += 20

        page_data['allowed', 'callers', 'lines'] = allowed
    else:
        page_data['allowed','show'] = False


    # If the responder has a target, draw a target line on the page
    if r_info.target_ident or r_info.target_ident_required:
        page_data['targetline','show'] = True
    else:
        page_data['targetline','show'] = False

    # normally no output ellipse is shown
    page_data['output','show'] = False


    # fill in the box regarding this responder
    if pageinfo.restricted:
        page_data['responder', 'responderaccess', 'text'] = "Restricted access"
    else:
        page_data['responder', 'responderaccess', 'text'] = "Open access"
    if label_list:
        page_data['responder', 'responderlabels', 'text'] = "Label: " + ','.join(label_list)
    else:
        page_data['responder', 'responderlabels', 'show'] = False
    page_data['responder', 'respondertype', 'text'] = "Responder: " + r_info.responder
    page_data['responder', 'responderbrief', 'text'] = pageinfo.brief

    # submit_data information
    if r_info.submit_option or r_info.submit_required:
        page_data['submitdata','show'] = True
        if r_info.submit_list:
            s_list = []
            s = 0
            for item in r_info.submit_list:
                s_list.append([0,s,item])
                s += 20
            page_data['submitdata', 'submitlist','lines'] = s_list
        # show the return value
        if r_info.responder == "ColourSubstitute":
            page_data['submitdata','submitdatareturn','text'] = "Returns a dictionary of strings: colour strings"
        elif r_info.responder == "SetCookies":
            page_data['submitdata','submitdatareturn','text'] = "Returns an instance of http.cookies.BaseCookie"
        elif r_info.responder == "GetDictionaryDefaults":
            page_data['submitdata','submitdatareturn','text'] = "Returns a dictionary with default values"
        elif r_info.responder == "SubmitJSON":
            page_data['submitdata','submitdatareturn','text'] = "Returns a dictionary"
            # no target, but include a target line
            page_data['targetline','show'] = True
            # change 'Target Page' to 'Output'
            page_data['submitdata','output', 'text'] = "Output"
            # show an output ellipse
            page_data['output','show'] = True
            page_data['output','textout', 'text'] = "Send JSON data"
            page_data['output','textout', 'x'] = 320
        elif r_info.responder == "SubmitPlainText":
            page_data['submitdata','submitdatareturn','text'] = "Returns a string"
            # no target, but include a target line
            page_data['targetline','show'] = True
            # change 'Target Page' to 'Output'
            page_data['submitdata','output', 'text'] = "Output"
            # show an output ellipse
            page_data['output','show'] = True
            page_data['output','textout', 'text'] = "Send plain text"
            page_data['output','textout', 'x'] = 320
        elif r_info.responder == "SubmitCSS":
            page_data['submitdata','submitdatareturn','text'] = "Returns a style"
            # no target, but include a target line
            page_data['targetline','show'] = True
            # change 'Target Page' to 'Output'
            page_data['submitdata','output', 'text'] = "Output"
            # show an output ellipse
            page_data['output','show'] = True
            page_data['output','textout', 'text'] = "Send CSS data"
            page_data['output','textout', 'x'] = 320
        elif r_info.responder == "MediaQuery":
            page_data['submitdata','submitdatareturn','text'] = "Returns a dictionary of media queries : CSS targets"
            # no target, but include a target line
            page_data['targetline','show'] = True
            # change 'Target Page' to 'Output'
            page_data['submitdata','output', 'text'] = "Output"
            # show an output ellipse
            page_data['output','show'] = True
            page_data['output','textout', 'text'] = "Update query:target items"
            page_data['output','textout', 'x'] = 320
        elif r_info.responder == "SubmitIterator":
            page_data['submitdata','submitdatareturn','text'] = "Returns a binary file iterator"
            # no target, but include a target line
            page_data['targetline','show'] = True
            # change 'Target Page' to 'Output'
            page_data['submitdata','output', 'text'] = "Output"
            # show an output ellipse
            page_data['output','show'] = True
            page_data['output','textout', 'text'] = "Send Binary data"
            page_data['output','textout', 'x'] = 320

        # show the fail page
        _show_submit_data_failpage(project, page_data, r_info)
    else:
        page_data['submitdata','show'] = False
        page_data['submitdata_failpage','show'] = False

    # The target page
    _show_target(project, page_data, r_info)

    # validation option
    _show_validate_fail(project, page_data, r_info)

    # The alternate option
    _show_alternate(project, page_data, r_info)

    if r_info.responder == 'CaseSwitch':
        _show_caseswitch(project, page_data, r_info)
    elif r_info.responder == 'EmptyCallDataGoto':
        _show_emptycalldatagoto(project, page_data, r_info)
    elif r_info.responder == 'EmptyGoto':
        _show_emptygoto(project, page_data, r_info)
    elif r_info.responder == "MediaQuery":
        _show_mediaquery(project, page_data, r_info)


def _show_target(project, page_data, r_info):
    "The responder passes the call to this target"
    if r_info.target_ident or r_info.target_ident_required:
        page_data['target','show'] = True
        if r_info.target_ident:
            targetident = r_info.target_ident
            if isinstance(targetident, str):
                targetident = skilift.ident_from_label(project, targetident)
            if targetident is None:
                page_data['target','show'] = False
            elif isinstance(targetident, str):
                page_data['target', 'responderid', 'text'] = targetident
            elif isinstance(targetident, tuple) and (len(targetident) == 2):
                targetinfo = skilift.page_info(*targetident)
                if targetident[0] == project:
                    page_data['target', 'responderid', 'text'] = "Ident: " + str(targetident[1])
                else:
                    page_data['target', 'responderid', 'text'] = "Ident: " + targetident[0] + ", " + str(targetident[1])
                if targetinfo.restricted:
                    page_data['target', 'responderaccess', 'text'] = "Restricted access"
                else:
                    page_data['target', 'responderaccess', 'text'] = "Open access"
                if isinstance(r_info.target_ident, str):
                    page_data['target', 'responderlabels', 'text'] = "Targeted from responder as: " + r_info.target_ident
                else:
                    page_data['target', 'responderlabels', 'text'] = "Targeted from responder as: " + r_info.target_ident[0] + ", " + str(r_info.target_ident[1])
                page_data['target', 'responderbrief', 'text'] = targetinfo.brief
                if targetinfo.item_type == "RespondPage":
                    page_data['target', 'respondertype', 'text'] = "Responder: " + targetinfo.responder
                else:
                    page_data['target', 'respondertype', 'text'] = targetinfo.item_type
    else:
        page_data['target','show'] = False


def _show_submit_data_failpage(project, page_data, r_info):
    "The responder calls submit data, which, if it raises a FailPage, calls this"
    if r_info.fail_ident:
        page_data['submitdata_failpage','show'] = True
        failident = r_info.fail_ident
        if isinstance(failident, str):
            failident = skilift.ident_from_label(project, failident)
        if failident is None:
            page_data['submitdata_failpage','show'] = False
        elif isinstance(failident, str):
            page_data['submitdata_failpage', 'responderid', 'text'] = failident
        elif isinstance(failident, tuple) and (len(failident) == 2):
            failinfo = skilift.page_info(*failident)
            if failident[0] == project:
                page_data['submitdata_failpage', 'responderid', 'text'] = "Ident: " + str(failident[1])
            else:
                page_data['submitdata_failpage', 'responderid', 'text'] = "Ident: " + failident[0] + ", " + str(failident[1])
            if failinfo.restricted:
                page_data['submitdata_failpage', 'responderaccess', 'text'] = "Restricted access"
            else:
                page_data['submitdata_failpage', 'responderaccess', 'text'] = "Open access"
            if isinstance(r_info.fail_ident, str):
                page_data['submitdata_failpage', 'responderlabels', 'text'] = "Set in responder as: " + r_info.fail_ident
            else:
                page_data['submitdata_failpage', 'responderlabels', 'text'] = "Set in responder as: " + r_info.fail_ident[0] + ", " + str(r_info.fail_ident[1])
            page_data['submitdata_failpage', 'responderbrief', 'text'] = failinfo.brief
            if failinfo.item_type == "RespondPage":
                page_data['submitdata_failpage', 'respondertype', 'text'] = "Responder: " + failinfo.responder
            else:
                page_data['submitdata_failpage', 'respondertype', 'text'] = failinfo.item_type
    else:
        page_data['submitdata_failpage','show'] = False


def _show_validate_fail(project, page_data, r_info):
    "The responder validates received data, on failure calls this"

    if r_info.validate_option:
        page_data['validate','show'] = True
    else:
        page_data['validate','show'] = False
        return

    if r_info.validate_fail_ident:
        failident = r_info.validate_fail_ident
        if isinstance(failident, str):
            failident = skilift.ident_from_label(project, failident)
        if isinstance(failident, str):
            page_data['validate', 'responderid', 'text'] = failident
        elif isinstance(failident, tuple) and (len(failident) == 2):
            failinfo = skilift.page_info(*failident)
            if failident[0] == project:
                page_data['validate', 'responderid', 'text'] = "Ident: " + str(failident[1])
            else:
                page_data['validate', 'responderid', 'text'] = "Ident: " + failident[0] + ", " + str(failident[1])
            if failinfo.restricted:
                page_data['validate', 'responderaccess', 'text'] = "Restricted access"
            else:
                page_data['validate', 'responderaccess', 'text'] = "Open access"
            if isinstance(r_info.fail_ident, str):
                page_data['validate', 'responderlabels', 'text'] = "Set in responder as: " + r_info.fail_ident
            else:
                page_data['validate', 'responderlabels', 'text'] = "Set in responder as: " + r_info.fail_ident[0] + ", " + str(r_info.fail_ident[1])
            page_data['validate', 'responderbrief', 'text'] = failinfo.brief
            if failinfo.item_type == "RespondPage":
                page_data['validate', 'respondertype', 'text'] = "Responder: " + failinfo.responder
            else:
                page_data['validate', 'respondertype', 'text'] = failinfo.item_type


def _show_alternate(project, page_data, r_info):
    "The alternate page"

    if r_info.alternate_ident:
        page_data['alternatebox','show'] = True
    else:
        page_data['alternatebox','show'] = False
        return

    if r_info.alternate_ident:
        altident = r_info.alternate_ident
        if isinstance(altident, str):
            altident = skilift.ident_from_label(project, altident)
        if isinstance(altident, str):
            page_data['alternatebox', 'responderid', 'text'] = altident
        elif isinstance(altident, tuple) and (len(altident) == 2):
            altinfo = skilift.page_info(*altident)
            if altident[0] == project:
                page_data['alternatebox', 'responderid', 'text'] = "Ident: " + str(altident[1])
            else:
                page_data['alternatebox', 'responderid', 'text'] = "Ident: " + altident[0] + ", " + str(altident[1])
            if altinfo.restricted:
                page_data['alternatebox', 'responderaccess', 'text'] = "Restricted access"
            else:
                page_data['alternatebox', 'responderaccess', 'text'] = "Open access"
            if isinstance(r_info.alternate_ident, str):
                page_data['alternatebox', 'responderlabels', 'text'] = "Set in responder as: " + r_info.alternate_ident
            else:
                page_data['alternatebox', 'responderlabels', 'text'] = "Set in responder as: " + r_info.alternate_ident[0] + ", " + str(r_info.alternate_ident[1])
            page_data['alternatebox', 'responderbrief', 'text'] = altinfo.brief
            if altinfo.item_type == "RespondPage":
                page_data['alternatebox', 'respondertype', 'text'] = "Responder: " + altinfo.responder
            else:
                page_data['alternatebox', 'respondertype', 'text'] = altinfo.item_type

    if r_info.responder == 'CaseSwitch':
        page_data['alternatebox', 'alttext', 'text'] = "Called if no match found"
    elif r_info.responder == 'EmptyCallDataGoto':
        page_data['alternatebox', 'alttext', 'text'] = "Called if skicall.call_data has key with value"
    elif r_info.responder == 'EmptyGoto':
        page_data['alternatebox', 'alttext', 'text'] = "Called if widgfield is present with a value"


def _show_caseswitch(project, page_data, r_info):

    page_data['textgroup', 'transform'] = 'translate(500,600)'

    if r_info.widgfield:
        text_title = """<text x="0" y="90">CaseSwitch on widgfield %s</text>""" % r_info.widgfield
    else:
        text_title = ''

    table_element = ''
    if r_info.field_values_list:
        for index, item in enumerate(r_info.field_values_list):
            table_element += _caseswitchtable(index, r_info.field_values_list)

    else:
        table_element = ''
    page_data['textgroup', 'text'] = text_title + table_element


def _caseswitchtable(index, field_values_list):
    y = 100 + 60*index
    return """
<rect height="60" style="fill:white;stroke-width:3;stroke:black" width="600" x="0" y="%s" />
<line x1="180" y1="%s" x2="180" y2="%s" style="stroke-width:3;stroke:black" />
<text x="20" y="%s">%s</text>
<text x="200" y="%s">%s</text>
""" % (y, y, y+60, y+30, field_values_list[index][0],y+30, field_values_list[index][1])


def _show_emptycalldatagoto(project, page_data, r_info):
    value = 'UNKNOWN'
    if r_info.single_field:
        value = r_info.single_field
    page_data['textgroup', 'transform'] = 'translate(750,700)'
    page_data['textgroup', 'text'] = """
<text x="0" y="0">Test skicall.call_data["%s"]</text>
<text x="0" y="60">Called if key not present, or has empty value.</text>
""" % (value,)


def _show_emptygoto(project, page_data, r_info):
    value = 'UNKNOWN'
    if r_info.widgfield:
        value = r_info.widgfield
    page_data['textgroup', 'transform'] = 'translate(750,700)'
    page_data['textgroup', 'text'] = """
<text x="0" y="0">Test widgfield %s</text>
<text x="0" y="60">Called if widgfield not present, or has empty value.</text>
""" % (value,)



def _show_mediaquery(project, page_data, r_info):
    page_data['textgroup', 'transform'] = 'translate(50,550)'

    if r_info.field_values_list:
        text_title = """<line x1="450" y1="70" x2="450" y2="100" style="stroke-width:3;stroke:black" />
<text x="0" y="90">Query:target</text>"""
    else:
        return
    table_element = ''
    for index, item in enumerate(r_info.field_values_list):
        table_element += _mediaquerytable(index, r_info.field_values_list)
    page_data['textgroup', 'text'] = text_title + table_element


def _mediaquerytable(index, field_values_list):
    y = 100 + 60*index
    return """
<rect height="60" style="fill:white;stroke-width:3;stroke:black" width="600" x="0" y="%s" />
<line x1="280" y1="%s" x2="280" y2="%s" style="stroke-width:3;stroke:black" />
<text x="20" y="%s">%s</text>
<text x="300" y="%s">%s</text>
""" % (y, y, y+60, y+30, field_values_list[index][0],y+30, field_values_list[index][1])







"""Contains widgets for inputting data"""

from urllib.parse import quote

from .. import skiboot, tag, excepts
from . import Widget, ClosedWidget, FieldArg, FieldArgList, FieldArgTable, FieldArgDict



class InputTable5(Widget):
    """Defines a div of four columns, the first just being label text, the second being text input fields,
       the third and fourth being submit buttons.  Each row of the table is a form, the
       form action ident is the same for all forms.
       On error the paragraph TextBlock is changed to the error message."""

    # This class does not display any error messages
    display_errors = False

    # js_validators is a class attribute, True if javascript validation is enabled
    js_validators=True

    arg_descriptions = {
                        'div_class':FieldArg("cssclass", ''),
                        'size':FieldArg("text", ''),
                        'maxlength':FieldArg("text", ''),
                        'required':FieldArg("boolean", False),
                        'action':FieldArg("url",''),
                        'button_text1':FieldArg("text", "Submit", valdt=True),
                        'button_text2':FieldArg("text", "Submit", valdt=True),
                        'button1_class':FieldArg("cssclass", ""),
                        'button2_class':FieldArg("cssclass", ""),
                        'col_label':FieldArgList('text'),
                        'label_class':FieldArg("cssclass", ""),
                        'label_style':FieldArg("cssstyle", ""),
                        'col_input':FieldArgList('text', valdt=True),
                        'hidden_field1':FieldArgList("text", valdt=True),
                        'hidden_field2':FieldArgList("text", valdt=True),
                        'hidden_field3':FieldArgList("text", valdt=True),
                        'hidden_field4':FieldArgList("text", valdt=True),
                        'input_accepted_class':FieldArg("cssclass", ''),
                        'input_class':FieldArg("cssclass", ''),
                        'input_style':FieldArg("cssstyle", ''),
                        'input_errored_class':FieldArg("cssclass", ''),
                        'set_input_accepted':FieldArgDict("boolean", jsonset=True),
                        'set_input_errored':FieldArgDict("boolean", jsonset=True)
                       }

    def __init__(self, name=None, brief='', **field_args):
        """
        div_class: the class attribute of each form div
        size: The number of characters appearing in each text input area
        maxlength: The maximum number of characters accepted in each text area
        required: Set True to put the 'required' flag into each input field
        action: The label or ident of the action page
        button_text1: text appearing on the first buttons - if empty, no button 1 is shown
        button_text2: text appearing on the second buttons - if empty, no button 2 is shown
        button1_class: class set on button1
        button2_class: class set on button2
        col_label: a list of all the text strings appearing in the first column
        label_class: the class to apply to each label
        col_input: a list of all the text strings appearing in the input fields
        hidden_field1-4: four hidden fields, each being a list
        input_accepted_class: A class which can be added on each input field
        input_class: A class which can be set on each input field
        input_errored_class: A class which can be added on each input field
        The following two dictionaries shoul have keys equal to the integer row number (starting
         at zero) and value True or False
        set_input_accepted: Dictionary of col index keys that should be set with input_accepted_class
        set_input_errored: Dictionary of col index keys that should be set with input_errored_class
        """
        Widget.__init__(self, name=name, tag_name="div", brief=brief, **field_args)


    def _build(self, page, ident_list, environ, call_data, lang):
        "Build the table"

        # get the widget ident
        ident_value = page.ident_data_string

        # get the form action url
        action_url = skiboot.get_url(self.get_field_value("action"), proj_ident=page.proj_ident)
        if not action_url:
            # setting self._error replaces the entire tag by the self._error message
            self._error = "Warning: broken link"
            return


        len_label = len(self.fields["col_label"])
        len_input = len(self.fields["col_input"])
        len_hidden_field1 = len(self.fields["hidden_field1"])
        len_hidden_field2 = len(self.fields["hidden_field2"])
        len_hidden_field3 = len(self.fields["hidden_field3"])
        len_hidden_field4 = len(self.fields["hidden_field4"])

        rows = max(len_label, len_input)
        if not rows:
            rows=1

        input_name = self.get_formname("col_input")
        button_name1 = self.get_formname("button_text1")
        button_value1 = self.get_field_value('button_text1')
        button1_class = self.get_field_value('button1_class')
        button_name2 = self.get_formname("button_text2")
        button_value2 = self.get_field_value('button_text2')
        button2_class = self.get_field_value('button2_class')
        hidden_field1_name = self.get_formname("hidden_field1")
        hidden_field2_name = self.get_formname("hidden_field2")
        hidden_field3_name = self.get_formname("hidden_field3")
        hidden_field4_name = self.get_formname("hidden_field4")

        label_class = self.get_field_value('label_class')
        label_style = self.get_field_value('label_style')

        form_id = self.get_id() + '_'

        if self.get_field_value('input_accepted_class') and self.get_field_value("set_input_accepted"):
            set_input_accepted = self.get_field_value("set_input_accepted")
        else:
            set_input_accepted = {}

        if self.get_field_value('input_errored_class') and self.get_field_value("set_input_errored"):
            set_input_errored = self.get_field_value("set_input_errored")
        else:
            set_input_errored = {}

        input_class = self.get_field_value('input_class')
        input_style = self.get_field_value('input_style')


        for rownumber in range(rows):
            if rownumber<len_label:
                col_label = self.get_field_value("col_label")[rownumber]
            else:
                col_label = ''
            if rownumber<len_input:
                col_input = self.get_field_value("col_input")[rownumber]
            else:
                col_input = ''
            if rownumber<len_hidden_field1:
                col_hidden_field1 = self.get_field_value("hidden_field1")[rownumber]
            else:
                col_hidden_field1 = ''
            if rownumber<len_hidden_field2:
                col_hidden_field2 = self.get_field_value("hidden_field2")[rownumber]
            else:
                col_hidden_field2 = ''
            if rownumber<len_hidden_field3:
                col_hidden_field3 = self.get_field_value("hidden_field3")[rownumber]
            else:
                col_hidden_field3 = ''
            if rownumber<len_hidden_field4:
                col_hidden_field4 = self.get_field_value("hidden_field4")[rownumber]
            else:
                col_hidden_field4 = ''

            # for every table row, create a form
            formrow = tag.Part(tag_name="form", attribs ={"method":"post", "action": action_url})

            # 1st column is label text 
            if label_class:
                formrow[0] = tag.Part(tag_name="label", attribs={'class':label_class})
            else:
                formrow[0] = tag.Part(tag_name="label")
            if label_style:
                formrow[0].update_attribs({"style":label_style})
            formrow[0][0] = col_label

            # 2nd column is a text input field
            formrow[1] = tag.ClosedPart(tag_name="input",
                                    attribs ={"name":input_name,
                                    "value":col_input,
                                    "type":"text"})
            if self.get_field_value('size'):
                formrow[1].update_attribs({"size":self.get_field_value('size')})
            if self.get_field_value('maxlength'):
                formrow[1].update_attribs({"maxlength":self.get_field_value('maxlength')})
            if self.get_field_value('required'):
                formrow[1].update_attribs({"required":"required"})
            # set an id in the input field for the 'label for' tag
            this_id = form_id + str(rownumber) + '_0_1'
            formrow[1].insert_id(this_id)
            # set the label 'for' attribute
            formrow[0].update_attribs({'for':formrow[1].get_id()})
            if (rownumber in set_input_errored) and set_input_errored[rownumber]:
                if input_class:
                    formrow[1].update_attribs({"class":input_class + ' ' + self.get_field_value('input_errored_class')})
                else:
                    formrow[1].update_attribs({"class":self.get_field_value('input_errored_class')})
            elif (rownumber in set_input_accepted) and set_input_accepted[rownumber]:
                if input_class:
                    formrow[1].update_attribs({"class":input_class + ' ' + self.get_field_value('input_accepted_class')})
                else:
                    formrow[1].update_attribs({"class":self.get_field_value('input_accepted_class')})
            elif input_class:
                formrow[1].update_attribs({"class":input_class})

            if input_style:
                formrow[1].update_attribs({"style":input_style})

            # 3rd column is a submit button
            if button_value1:
                formrow.append(tag.ClosedPart(tag_name="input",
                                 attribs ={"name":button_name1,
                                           "value":button_value1,
                                           "class":button1_class,
                                           "type":"submit"}))
            # 4th column is a submit button
            if button_value2:
                formrow.append(tag.ClosedPart(tag_name="input",
                                 attribs ={"name":button_name2,
                                           "value":button_value2,
                                           "class":button2_class,
                                           "type":"submit"}))

            # all submissions always have an 'ident' hidden field to provide the ident of the calling page
            formrow.append(tag.ClosedPart(tag_name="input",
                                          attribs ={"name":'ident',
                                                 "value":ident_value,
                                                 "type":"hidden"}))

            # hidden field on the form
            if col_hidden_field1:
                formrow.append(tag.ClosedPart(tag_name = "input",
                                            attribs =  {"name":hidden_field1_name,
                                                        "value":col_hidden_field1,
                                                        "type":"hidden"}))

            # Second hidden field on the form
            if col_hidden_field2:
                formrow.append(tag.ClosedPart(tag_name = "input",
                                            attribs =  {"name":hidden_field2_name,
                                                        "value":col_hidden_field2,
                                                        "type":"hidden"}))
            # Third hidden field on the form
            if col_hidden_field3:
                formrow.append(tag.ClosedPart(tag_name = "input",
                                            attribs =  {"name":hidden_field3_name,
                                                        "value":col_hidden_field3,
                                                        "type":"hidden"}))
            # Fourth hidden field on the form
            if col_hidden_field4:
                formrow.append(tag.ClosedPart(tag_name = "input",
                                            attribs =  {"name":hidden_field4_name,
                                                        "value":col_hidden_field4,
                                                        "type":"hidden"}))

            # place form in a div
            self[rownumber] = tag.Part(tag_name="div")
            if self.get_field_value('div_class'):
                self[rownumber].attribs = {"class": self.get_field_value('div_class')}
            self[rownumber][0] = formrow


    def _build_js(self, page, ident_list, environ, call_data, lang):
        """Sets a submit event handler"""
        jscript = """  $('#{ident}').find( 'form').each(function() {{
    $(this).on("submit", function(e) {{
        SKIPOLE.widgets['{ident}'].eventfunc(e);
        }});
    }});
""".format(ident=self.get_id())
        return jscript + self._make_fieldvalues('input_accepted_class', 'input_errored_class')

    @classmethod
    def description(cls):
        """Returns a text string to illustrate the widget"""
        return """
<div>   <!-- with widget id and class widget_class -->
  <!-- Then a div containing a form, label, input field, buttons and hidden fields, repeated for each list  item -->
  <div> <!-- with CSS class attribute set to div_class if a class is set -->
    <form method=\"post\"> <!-- action attribute set to action field -->
      <label> <!-- with col_label text and CSS class label_class -->
      </label >
        <input type="text" /> <!-- all with same CSS input_class and size and maxlength, with values from col_input -->
        <input type="submit" /> <!-- with value button_text1, CSS class button1_class and name widgfield derived from button_text1 field name -->
        <input type="submit" /> <!-- with value button_text2, CSS class button2_class and name widgfield derived from button_text2 field name -->
        <!-- hidden input fields -->
    </form>                              
  </div>
  <!--  The above repeated for each item in col_input or col_label, whichever is greater -->
</div>"""

